# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'dualscale.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_container(object):
    def setupUi(self, container):
        container.setObjectName("container")
        container.resize(638, 627)
        self.gridLayout = QtWidgets.QGridLayout(container)
        self.gridLayout.setObjectName("gridLayout")
        self.glayout = QtWidgets.QGridLayout()
        self.glayout.setObjectName("glayout")
        self.gridLayout.addLayout(self.glayout, 0, 0, 1, 1)

        self.retranslateUi(container)
        QtCore.QMetaObject.connectSlotsByName(container)

    def retranslateUi(self, container):
        _translate = QtCore.QCoreApplication.translate
        container.setWindowTitle(_translate("container", "Form"))



import itertools
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtProperty, Qt
import sys
from math import pi, isinf, sqrt, asin, ceil, cos, sin, floor, ceil

# See file COPYING in this source tree
__copyright__ = ('\n'.join([
    'Copyright 2008-2012 Meinert Jordan <meinert@gmx.at>',
    # Originally written in C++
    # http://qt-apps.org/content/show.php/QScale?content=148053

    'Copyright 2014, Fabián Inostroza',
    # Ported to PyQt4
    # http://pastebin.com/kzp7f7DS

    'Copyright 2017, EPC Power Corp.',
    # Ported to PyQt5
    # Max/min positions correct and vertical orientation can be flipped.
]))
__license__ = 'GPLv2+'


class QScale(QtWidgets.QWidget):

    def __init__(self, parent=None, in_designer=False):
        QtWidgets.QWidget.__init__(self, parent=parent)

        self.in_designer = in_designer

        self.m_minimum = 0 # Minimum value of scale; default value 0 
        self.m_maximum = 100 # Maximum value of scale; default value 100
        self.m_value = 0 # Current value (where needle would point); def val = 0
        self.m_paintMode = 0 # The paint mode, which determines how much of
                             # the scale is painted.
                             # 0 = All, 1 = color ranges, scale markers, labels,
                             # needle. 2 = only needle. 3 = needle, cover.

        self.m_labelsVisible = True # Determines if numbers of scale would show
        self.m_scaleVisible = True # Determines if markers of scale would show

        # Values that help represent values that make up the QScale widget.
        self.m_borderWidth = 6
        self.m_labelsFormat = 'g'
        self.m_labelsPrecision = -1
        self.m_majorStepSize = 0
        self.m_minorStepCount = 0

        self.m_invertedAppearance = False # This value is not currently used.
        self.m_orientations = QtCore.Qt.Horizontal | QtCore.Qt.Vertical

        self.vertically_flipped = False # New property specifying if flipped.

        self.setBackgroundRole(QtGui.QPalette.Base)

        self.labelSample = ""
        self.updateLabelSample()
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, 
                           QtWidgets.QSizePolicy.Expanding)

        self.breakpoints = []
        self.colors = []

        self.isBlue = False

    def setMinimumSize(self, width=None, height=None, painter=None):
        self.updateLabelSample()

        if painter is None:
            painter = QtGui.QPainter(self)

        # TODO: CAMPid 079370432832243267955437254329546425654321
        rect = painter.boundingRect(QtCore.QRectF(0, 0, self.width(), 
                                    self.height()), 
                                    QtCore.Qt.AlignBottom | 
                                    QtCore.Qt.AlignHCenter, self.labelSample)

        # TODO: CAMPid 07899789654211527951677432169
        # Specify if the widget is vertical.
        if ((not (self.m_orientations & QtCore.Qt.Vertical)) 
             ^ (not (self.m_orientations & QtCore.Qt.Horizontal))):
            vertical = self.m_orientations & QtCore.Qt.Vertical
        else:
            vertical = self.height() > self.width()

        wLabel = rect.width() if self.m_labelsVisible else 0
        hLabel = rect.height() if self.m_labelsVisible else 0

        if vertical:
            wLabel, hLabel = hLabel, wLabel

        if width is None:
            width = ceil(2 * self.m_borderWidth + wLabel + 1)
        if height is None:
            height = ceil(2 * (1 + self.m_borderWidth + hLabel))

        if vertical:
            height, width = width, height

        super().setMinimumSize(width, height)

    def setMinimum(self, max):
        if not isinf(max):
            self.m_maximum = max
        self.updateLabelSample()
        self.update()

    def maximum(self):
        return self.m_maximum

    def setRange(self, min, max):
        if not isinf(min):
            self.m_minimum = min
        if not isinf(max):
            self.m_maximum = max
        self.updateLabelSample()
        self.update()

    def setColorRanges(self, colors, breakpoints):
        if len(colors) == 0:
            # TODO: something better
            raise Exception('no colors')

        if not all(x < y for x, y in zip(breakpoints, breakpoints[1:])):
            # TODO: something better
            raise ValueError('Monotonicity')

        if len(colors) - len(breakpoints) != 1:
            # TODO: something better
            raise ValueError('Bad set of color range lists')

        self.breakpoints = breakpoints
        self.colors = colors

    def setValue(self, val):
        self.m_value = val
        self.update()

    def value(self):
        return self.m_value

    def setLabelsVisible(self, visible):
        self.m_labelsVisible = visible
        self.update()

    def isLabelsVisible(self):
        return self.m_labelsVisible

    def setScaleVisible(self, visible):
        self.m_scaleVisible = visible
        self.update()

    def isScaleVisible(self):
        return self.m_scaleVisible

    def setBorderWidth(self, width):
        self.m_borderWidth = width if width > 0 else 0
        self.update()

    def borderWidth(self):
        return self.m_borderWidth

    def setLabelsFormat(self, fmt, precision):
        self.m_labelsFormat = fmt
        self.m_labelsPrecision = precision
        self.updateLabelSample()
        self.update()

    def setMajorStepSize(self, stepsize):
        self.m_majorStepSize = stepsize
        self.update()

    def majorStepSize(self):
        return self.m_majorStepSize

    def setMinorStepSize(self, stepcount):
        self.m_minorStepCount = stepcount
        self.update()

    def minorStepCount(self):
        return self.m_minorStepCount

    def setInvertedAppearance(self, invert):
        self.m_invertedAppearance = invert
        self.update()

    def invertedAppearance(self):
        return self.m_invertedAppearance

    # orientations es Qt::Orientations
    def setOrientations(self, orientations):
        self.m_orientations = orientations
        self.update()

    def orientations(self):
        return self.m_orientations

    def resizeEvent(self, re):
        super(QScale,self).resizeEvent(re)

    def paintEvent(self, paintEvent):

        # Set the values.
        
        painter = QtGui.QPainter(self)

        painter.setRenderHint(QtGui.QPainter.Antialiasing,True)

        self.setMinimumSize(painter=painter)

        # TODO: CAMPid 07899789654211527951677432169

        # Determine the criterion of what makes the widget vertical.
        if ((not (self.m_orientations & QtCore.Qt.Vertical)) 
            ^ (not (self.m_orientations & QtCore.Qt.Horizontal))):
            vertical = self.m_orientations & QtCore.Qt.Vertical
        else:
            vertical = self.height() > self.width()

        # Acquire widget width and height values.
        wWidget = self.width()
        hWidget = self.height()

        # TODO: CAMPid 079370432832243267955437254329546425654321
        # Acquire the height and width of the individual labels.
        boundingRect = painter.boundingRect(QtCore.QRectF(0, 0, self.width(), 
                                            self.height()), 
                                            QtCore.Qt.AlignBottom | 
                                            QtCore.Qt.AlignHCenter,
                                            self.labelSample)

        # Set the width and height of the labels.
        wLabel = boundingRect.width() if self.m_labelsVisible else 0
        hLabel = boundingRect.height() if self.m_labelsVisible else 0

        # Swap width and height values if the widget is vertical.
        if vertical:
            wWidget, hWidget = hWidget, wWidget
            wLabel, hLabel = hLabel, wLabel


        # Get the width and height of the scale itself 
        # (based on the label and actual widget values)
        wScale = wWidget - wLabel - 2.0 * self.m_borderWidth

        hScale = 0.5 * hWidget - hLabel - self.m_borderWidth

        # Acquire radius of drawn scale based off of width and height of scale.
        radius = 0.125 * wScale ** 2 / hScale + 0.5 * hScale

        # Acquire the center point and modify radius if needed.
        if radius < hScale + 0.5 * hWidget - self.m_borderWidth:
            radius = ((4.0 * (hLabel + self.m_borderWidth)
                       + sqrt(4.0 * (hLabel + self.m_borderWidth) ** 2
                       + 3.0 * wScale ** 2)) / 3.0
                       - hLabel - 2.0 * self.m_borderWidth)
            center = QtCore.QPointF(0.5 * wWidget, hWidget 
                                    - self.m_borderWidth)
        else:
            center = QtCore.QPointF(0.5 * wWidget, 
                                    radius + hLabel + self.m_borderWidth)

        # Calculate where certain things start and their span.
        angleSpan = -360.0 / pi * asin(wScale / (2.0 * radius))
        angleStart = 90.0 - 0.5 * angleSpan

        valueSpan = self.m_maximum - self.m_minimum

        # Calculate the size of the tick marks.
        majorStep = abs(valueSpan) * self.max(wLabel, 1.5 
                    * boundingRect.height()) / wScale
        order = 0

        while majorStep < 1:
            majorStep *= 10
            order -= 1

        while majorStep >= 10:
            majorStep /= 10
            order += 1

        if majorStep > 5:
            majorStep = 10 * 10 ** order
            minorSteps = 5
        elif majorStep > 2:
            majorStep = 5 * 10 ** order
            minorSteps = 5
        else:
            majorStep = 2 * 10 ** order
            minorSteps = 4

        if self.m_majorStepSize > 0:
            majorStep = self.m_majorStepSize
        if self.m_minorStepCount > 0:
            minorSteps = self.m_minorStepCount

        scaleWidth = self.min(self.min(0.25 * (hWidget - self.m_borderWidth), 
                                       0.25 * radius), 2.5 
                                       * boundingRect.height())
        minorScaleWidth = scaleWidth * 0.4

        # Change the orientation of the scale if the scale is vertical.
        if vertical:
            painter.rotate(90)
            painter.translate(0, -hWidget + wLabel / 4.0)

        def drawColorRanges(self):

            # Make starting point be top left corner instead of top right corner
            if vertical and self.vertically_flipped:
                # Weird. Seems to get the position as if the angle is 0 degrees
                painter.translate(0, painter.transform().dx())
                # Correctly repositions color scale.
                painter.translate(0, -wLabel / 4.0)

            if len(self.colors) > 0:
                transform = painter.transform()
                valueSpan = self.m_maximum - self.m_minimum
                rangeValueStart = self.m_minimum

                for breakpoint, color in itertools.zip_longest(
                                                self.breakpoints, self.colors):
                    # Consider color for range [rangeValueStart, breakpoint]
                    if breakpoint is None or breakpoint > rangeValueStart:
                        if rangeValueStart < self.m_maximum:
                            if not vertical or self.vertically_flipped:
                                rangeAngleStart = (angleStart + angleSpan 
                                                   * (rangeValueStart 
                                                   - self.m_minimum) 
                                                   / valueSpan)
                            elif vertical and not self.vertically_flipped:
                                if breakpoint is None:
                                    breakpoint = self.m_maximum
                                rangeAngleStart = (angleStart + angleSpan 
                                                   * (self.m_maximum 
                                                   - breakpoint) / valueSpan)
                            try:
                                if not vertical or self.vertically_flipped:
                                    rangeAngleEnd = (angleStart + angleSpan 
                                                     * (breakpoint 
                                                     - self.m_minimum) 
                                                     / valueSpan)
                                elif vertical and not self.vertically_flipped:
                                    rangeAngleEnd = (angleStart + angleSpan 
                                                     * (self.m_maximum 
                                                     - rangeValueStart) 
                                                     / valueSpan)
                            except TypeError:
                                rangeAngleEnd = angleStart + angleSpan
                            # max because of angles going counter clockwise...
                            rangeAngleEnd = max(rangeAngleEnd, angleStart 
                                                + angleSpan)
                            rangeAngleSpan = rangeAngleEnd - rangeAngleStart

                            if vertical and self.vertically_flipped:
                                rangeAngleStart += 180
                                rangeAngleEnd += 180

                            painter.setPen(color)
                            painter.setBrush(color)
                            qpp = QtGui.QPainterPath()
                            r = radius - 0.8 * scaleWidth
                            d = 2 * r
                            x = center.x() - r
                            y = center.y() - r                           


                            y_offset = y
                            # Color scale orientation changed if flipped.
                            # Keep in mind, it isn't a PERFECT flip.
                            if not (vertical and self.vertically_flipped):
                                y_offset = y
                            else:
                                y_offset = -y -d
                            qpp.arcMoveTo(x, y_offset, d, d, rangeAngleStart)
                            qpp.arcTo(x, y_offset, d, d, rangeAngleStart, 
                                      rangeAngleSpan)

                            outer = QtGui.QPainterPath()
                            r = radius - 0.6 * scaleWidth
                            d = 2 * r
                            x = center.x() - r
                            y = center.y() - r

                            # Color scale orientation changed if flipped.
                            if not (vertical and self.vertically_flipped):
                                y_offset = y
                            else:
                                y_offset = -y -d

                            outer.arcMoveTo(x, y_offset, d, d, rangeAngleStart 
                                            + rangeAngleSpan)
                            outer.arcTo(x, y_offset, d, d, rangeAngleStart 
                                        + rangeAngleSpan, -rangeAngleSpan)

                            qpp.connectPath(outer)
                            qpp.closeSubpath()
                            painter.drawPath(qpp)
                            painter.setTransform(transform)

                            rangeValueStart = breakpoint

            painter.resetTransform()

        def drawScaleMarkers(self):
            painter.setPen(QtGui.QPen(self.palette().color(QtGui.QPalette.Text), 
                           1))

            # Only draw if the scale is visible and enough space for tick marks.
            if self.m_scaleVisible and majorStep != 0:

                # Rotate the painter to accomodate for vertical scale.
                if vertical:
                    if not self.vertically_flipped:
                        painter.rotate(90)
                        painter.translate(0, -hWidget + wLabel / 4.0)
                    else:
                        painter.rotate(90) 
                        painter.translate(0, -self.width())
                        painter.translate(0, -(-hWidget + wLabel / 4.0)) 

                # Account for flipping vertical scales.
                if vertical and self.vertically_flipped:
                    painter.translate(center.x(), -center.y())
                else:
                    painter.translate(center) 

                # Turn 180 degrees to account for flipped vertical scales.
                if vertical and self.vertically_flipped:
                    painter.rotate(self.m_minimum % ceil(float(majorStep) 
                                   / float(minorSteps)) / float(valueSpan) 
                                   * angleSpan - angleStart + 180)
                else:
                    painter.rotate(self.m_minimum % ceil(float(majorStep) 
                                   / float(minorSteps)) / float(valueSpan) 
                                   * angleSpan - angleStart)

                offsetCount = (minorSteps - ceil(self.m_minimum % majorStep) 
                               / float(majorStep) * minorSteps) % minorSteps

                # Actual drawing of tick marks done here.
                for i in range(0, floor(minorSteps * abs(valueSpan) / majorStep) 
                               + 1):
                    if i % minorSteps == offsetCount:
                        # Draw bigger line for like every 5 or 10 or so.
                        painter.drawLine(QtCore.QLineF(radius - scaleWidth, 0, 
                                         radius, 0))
                    else:
                        # Draw smaller line for the other tick marks.
                        painter.drawLine(QtCore.QLineF(radius - scaleWidth, 0,
                                                       radius - minorScaleWidth, 
                                                       0))
                    # Rotate the painter a bit for the next tick mark.
                    painter.rotate(majorStep * angleSpan 
                                   / (-abs(valueSpan) * minorSteps))

                painter.resetTransform()

        def drawLabels(self):

            # Draw the numbers on the scale.
            if self.m_labelsVisible and majorStep != 0:
                # x represents the amount tick marks on the scale.
                x = range(int(ceil(self.min(self.m_minimum, self.m_maximum) 
                                   / majorStep)), 
                                  int(self.max(self.m_minimum, self.m_maximum) 
                                   / majorStep) + 1)

                # Prep each tick mark, draw right value and at right positioning
                for i in x:
                    # Get the particular angle the number should be drawn at.
                    u = pi / 180.0 * ((majorStep * i - self.m_minimum) 
                           / float(valueSpan) * angleSpan + angleStart)
                    position = QtCore.QRect()

                    # Acquire the position of the number based on orientation.
                    if vertical:
                        if not self.vertically_flipped:
                            align = QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter
                            position = QtCore.QRect(self.width() - center.y() 
                                                    + radius * sin(u), 0,
                                                    self.width(), self.height() 
                                                    + 2 * radius * cos(u))
                        else:
                            align = (QtCore.Qt.AlignRight 
                                     | QtCore.Qt.AlignVCenter)
                            position = QtCore.QRect(center.y() - radius * sin(u) 
                                                    - self.width(), 0,
                                                    self.width(), self.height() 
                                                    + 2 * radius * cos(u))
                    else:
                        align = QtCore.Qt.AlignHCenter | QtCore.Qt.AlignBottom
                        position = QtCore.QRect(0, 0, 2.0 * (center.x() + radius
                                                * cos(u)),
                                                center.y() - radius * sin(u))
                    painter.resetTransform()
                    # TODO: add usage of m_labelsFormat and m_labelsPrecision

                    # Draw the number in an order that depends if the scale is
                    # vertical or not.
                    if vertical:
                        painter.drawText(position, align, 
                                        '{}'.format((x.stop + x.start - i - 1) 
                                                     * majorStep))
                    else:
                        painter.drawText(position, align, 
                                         '{}'.format(i * majorStep))

        def drawNeedle(self):
            painter.resetTransform()
            # CHange the painter's orientation depending on scale's orientation.
            if vertical:
                if not self.vertically_flipped:
                    painter.rotate(90) 
                    painter.translate(0, -hWidget + wLabel / 4.0) 
                else:
                    painter.rotate(90) 
                    painter.translate(0, -self.width())
                    painter.translate(0, -(-hWidget + wLabel / 4.0)) 

            if vertical and self.vertically_flipped:
                painter.translate(center.x(), -center.y())
            else:
                painter.translate(center) 

            # Get the calibration of the needle.
            if vertical:
                if not self.vertically_flipped:
                    painter.rotate((-self.m_maximum + self.m_value) 
                                    / float(valueSpan) * angleSpan - angleStart) 
                else:
                    painter.rotate((self.m_minimum - self.m_value) 
                                    / float(valueSpan) * angleSpan - angleStart) 
                    painter.rotate(180) 
            else:
                painter.rotate((self.m_minimum - self.m_value) 
                                / float(valueSpan) * angleSpan - angleStart) 

            # Actually draw the needle based off of the previous settings set up 
            painter.setPen(QtCore.Qt.NoPen)
            painter.setBrush(self.palette().color(QtGui.QPalette.Text))
            self.polygon = QtGui.QPolygon()

            # python does not need the first parameter (number of points)
            self.polygon.setPoints(0, -2, int(radius) - 10, -2 , int(radius), 0,
                              int(radius) - 10, 2, 0, 2) 

            painter.drawConvexPolygon(self.polygon)

            if not self.isBlue:
                painter.setPen(QtGui.QPen(self.palette().
                                                color(QtGui.QPalette.Base), 2)) 
            else:
                painter.setPen(QtGui.QPen(Qt.blue))
            painter.drawLine(0, 0, radius - 15, 0) 
            painter.resetTransform()

        def drawCover(self, center):

            painter.setPen(QtCore.Qt.NoPen)
            painter.setBrush(self.palette().color(QtGui.QPalette.Mid))

            # Draw rectangle component depending on the orientation of scale.
            # The circular part of scale also varies on orientation of scale.
            if vertical:
                if self.vertically_flipped:
                    painter.drawRect(QtCore.QRect(self.width() 
                                                  - self.m_borderWidth, 0, 
                                                  self.m_borderWidth, 
                                                  self.height()))
                else:
                    painter.drawRect(QtCore.QRect(0, 0, self.m_borderWidth, 
                                                  self.height()))
                center = QtCore.QPoint(self.width() - center.y() - wLabel / 4.0, 
                                       0.5 * self.height())
                u = 0.25 * (hWidget - wLabel) - center.x() - self.m_borderWidth
                if self.vertically_flipped:
                    center = QtCore.QPoint(-center.x() + self.width(), 
                                           center.y())
            else:
                pass
                painter.drawRect(QtCore.QRect(0, hWidget, wWidget, 
                                             -self.m_borderWidth))
                u = center.y() - self.m_borderWidth - 0.75 * hWidget

            # Draw the circular part of scale. It will be a very small portion
            # of a very large ellipse for the most part.
            u = self.max(u, 0.25 * radius)
            u = min(u, (radius - scaleWidth) - minorScaleWidth)
            painter.drawEllipse(center, u, u) 

        if self.m_paintMode == 0 or self.m_paintMode == 1:
            drawColorRanges(self)
            drawScaleMarkers(self)
            drawLabels(self)
        drawNeedle(self)
        if self.m_paintMode == 0 or self.m_paintMode == 3:
            drawCover(self, center)



    def updateLabelSample(self):
        margin = self.max(abs(self.m_minimum), abs(self.m_maximum))
        if self.min(self.m_minimum, self.m_maximum) < 0:
            wildcard = float(-8)
        else:
            wildcard = float(8)

        while margin < 1:
            margin *= 10
            wildcard /= 10

        while margin >= 10:
            margin /= 10
            wildcard *= 10

        # self.labelSample = QtCore.QString.number(wildcard, 
        #                        self.m_labelsFormat, self.m_labelsPrecision)
        # TODO: add usage of m_labelsFormat and m_labelsPrecision
        self.labelSample = '{}'.format(wildcard)

    def max(self, val1, val2):
        return val1 if val1 > val2 else val2

    def min(self, val1, val2):
        return val1 if val1 < val2 else val2

    # New functions help determining flip status.
    @pyqtProperty(bool)
    def flipped(self):
        return self.vertically_flipped;

    @flipped.setter
    def flipped(self, value):
        self.vertically_flipped = value


if __name__ == '__main__':
    global j
    j = 100
    def update():
        global j
        if j == 0:
            j = 100
        else:
            j -= 1
        scale.setValue(j)

    app = QtGui.QApplication(sys.argv)
    scale = QScale()
    timer = QtCore.QTimer()
    timer.setInterval(100)
    timer.timeout.connect(update)
    timer.start()
    scale.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'numberpad.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(832, 363)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.edit = QtWidgets.QLineEdit(Form)
        self.edit.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.edit.setObjectName("edit")
        self.gridLayout.addWidget(self.edit, 1, 2, 1, 4)
        self.button_decimal = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_decimal.sizePolicy().hasHeightForWidth())
        self.button_decimal.setSizePolicy(sizePolicy)
        self.button_decimal.setObjectName("button_decimal")
        self.gridLayout.addWidget(self.button_decimal, 4, 5, 1, 1)
        self.button_0 = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_0.sizePolicy().hasHeightForWidth())
        self.button_0.setSizePolicy(sizePolicy)
        self.button_0.setObjectName("button_0")
        self.gridLayout.addWidget(self.button_0, 5, 5, 1, 1)
        self.accept_button = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.accept_button.sizePolicy().hasHeightForWidth())
        self.accept_button.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("FontAwesome")
        self.accept_button.setFont(font)
        self.accept_button.setObjectName("accept_button")
        self.gridLayout.addWidget(self.accept_button, 3, 6, 1, 1)
        self.button_backspace = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_backspace.sizePolicy().hasHeightForWidth())
        self.button_backspace.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("FontAwesome")
        self.button_backspace.setFont(font)
        self.button_backspace.setObjectName("button_backspace")
        self.gridLayout.addWidget(self.button_backspace, 1, 6, 1, 1)
        self.button_1 = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_1.sizePolicy().hasHeightForWidth())
        self.button_1.setSizePolicy(sizePolicy)
        self.button_1.setObjectName("button_1")
        self.gridLayout.addWidget(self.button_1, 3, 2, 1, 1)
        self.button_2 = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_2.sizePolicy().hasHeightForWidth())
        self.button_2.setSizePolicy(sizePolicy)
        self.button_2.setObjectName("button_2")
        self.gridLayout.addWidget(self.button_2, 3, 3, 1, 1)
        self.button_6 = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_6.sizePolicy().hasHeightForWidth())
        self.button_6.setSizePolicy(sizePolicy)
        self.button_6.setObjectName("button_6")
        self.gridLayout.addWidget(self.button_6, 4, 4, 1, 1)
        self.button_4 = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_4.sizePolicy().hasHeightForWidth())
        self.button_4.setSizePolicy(sizePolicy)
        self.button_4.setObjectName("button_4")
        self.gridLayout.addWidget(self.button_4, 4, 2, 1, 1)
        self.button_3 = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_3.sizePolicy().hasHeightForWidth())
        self.button_3.setSizePolicy(sizePolicy)
        self.button_3.setObjectName("button_3")
        self.gridLayout.addWidget(self.button_3, 3, 4, 1, 1)
        self.label = QtWidgets.QLabel(Form)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 1, 1, 6)
        self.button_8 = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_8.sizePolicy().hasHeightForWidth())
        self.button_8.setSizePolicy(sizePolicy)
        self.button_8.setObjectName("button_8")
        self.gridLayout.addWidget(self.button_8, 5, 3, 1, 1)
        self.up_button = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.up_button.sizePolicy().hasHeightForWidth())
        self.up_button.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("FontAwesome")
        self.up_button.setFont(font)
        self.up_button.setObjectName("up_button")
        self.gridLayout.addWidget(self.up_button, 1, 1, 1, 1)
        self.button_7 = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_7.sizePolicy().hasHeightForWidth())
        self.button_7.setSizePolicy(sizePolicy)
        self.button_7.setObjectName("button_7")
        self.gridLayout.addWidget(self.button_7, 5, 2, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 5, 7, 1, 1)
        self.button_5 = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_5.sizePolicy().hasHeightForWidth())
        self.button_5.setSizePolicy(sizePolicy)
        self.button_5.setObjectName("button_5")
        self.gridLayout.addWidget(self.button_5, 4, 3, 1, 1)
        self.button_9 = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_9.sizePolicy().hasHeightForWidth())
        self.button_9.setSizePolicy(sizePolicy)
        self.button_9.setObjectName("button_9")
        self.gridLayout.addWidget(self.button_9, 5, 4, 1, 1)
        self.cancel_button = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cancel_button.sizePolicy().hasHeightForWidth())
        self.cancel_button.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("FontAwesome")
        self.cancel_button.setFont(font)
        self.cancel_button.setObjectName("cancel_button")
        self.gridLayout.addWidget(self.cancel_button, 5, 6, 1, 1)
        self.line = QtWidgets.QFrame(Form)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 2, 2, 1, 4)
        spacerItem1 = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 5, 0, 1, 1)
        self.gridLayout.setColumnStretch(0, 1)
        self.gridLayout.setColumnStretch(1, 1)
        self.gridLayout.setColumnStretch(2, 1)
        self.gridLayout.setColumnStretch(3, 1)
        self.gridLayout.setColumnStretch(4, 1)
        self.gridLayout.setColumnStretch(5, 1)
        self.gridLayout.setColumnStretch(6, 1)
        self.gridLayout.setColumnStretch(7, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.button_decimal.setText(_translate("Form", "."))
        self.button_0.setText(_translate("Form", "0"))
        self.accept_button.setText(_translate("Form", ""))
        self.button_backspace.setText(_translate("Form", ""))
        self.button_1.setText(_translate("Form", "1"))
        self.button_2.setText(_translate("Form", "2"))
        self.button_6.setText(_translate("Form", "6"))
        self.button_4.setText(_translate("Form", "4"))
        self.button_3.setText(_translate("Form", "3"))
        self.label.setText(_translate("Form", "TextLabel"))
        self.button_8.setText(_translate("Form", "8"))
        self.up_button.setText(_translate("Form", ""))
        self.button_7.setText(_translate("Form", "7"))
        self.button_5.setText(_translate("Form", "5"))
        self.button_9.setText(_translate("Form", "9"))
        self.cancel_button.setText(_translate("Form", ""))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'parameteredit.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(558, 504)
        Form.setProperty("style_small", True)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 3, 0, 1, 1)
        self.line = QtWidgets.QFrame(Form)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 2, 0, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem1, 6, 0, 1, 1)
        self.description = QtWidgets.QLabel(Form)
        self.description.setWordWrap(True)
        self.description.setObjectName("description")
        self.gridLayout.addWidget(self.description, 4, 0, 1, 1)
        self.to_device = Epc(Form)
        self.to_device.setProperty("label_visible", False)
        self.to_device.setProperty("tx", True)
        self.to_device.setProperty("show_enumeration_value", False)
        self.to_device.setObjectName("to_device")
        self.gridLayout.addWidget(self.to_device, 5, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.esc_button = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.esc_button.sizePolicy().hasHeightForWidth())
        self.esc_button.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("FontAwesome")
        self.esc_button.setFont(font)
        self.esc_button.setObjectName("esc_button")
        self.horizontalLayout.addWidget(self.esc_button)
        self.save_to_nv_button = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.save_to_nv_button.sizePolicy().hasHeightForWidth())
        self.save_to_nv_button.setSizePolicy(sizePolicy)
        self.save_to_nv_button.setObjectName("save_to_nv_button")
        self.horizontalLayout.addWidget(self.save_to_nv_button)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.description.setText(_translate("Form", "TextLabel"))
        self.esc_button.setText(_translate("Form", ""))
        self.save_to_nv_button.setText(_translate("Form", "Save All Parameters To NV"))
from epyqlib.widgets.epc import Epc



import can
import canmatrix.canmatrix
import epyqlib.pyqabstractitemmodel
from epyqlib.abstractcolumns import AbstractColumns
import epyqlib.canneo
from epyqlib.treenode import TreeNode
from PyQt5.QtCore import (Qt, QVariant, QModelIndex, pyqtSignal, QTimer)

# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class SignalNode(epyqlib.canneo.Signal, TreeNode):
    def __init__(self, signal, frame, tx=False, connect=None, tree_parent=None, parent=None):
        epyqlib.canneo.Signal.__init__(self, signal=signal, frame=frame, connect=connect, parent=parent)
        TreeNode.__init__(self, tx=tx, parent=tree_parent)

        self.fields = Columns(id=self.start_bit,
                              name=self.name,
                              length='{} b'.format(self.signal_size),
                              value='-',
                              dt='-',
                              count='')
        self.last_time = None

    def unique(self):
        # TODO: make it more unique
        return str(self.fields.id) + '__'

    def set_value(self, value, force=False, check_range=False):
        epyqlib.canneo.Signal.set_value(self,
                                        value=value,
                                        force=force,
                                        check_range=check_range)
        self.fields.value = self.full_string

    def set_data(self, data):
        try:
            self.set_human_value(data)
        except ValueError:
            raise
        else:
            self.frame.update_from_signals()


class MessageNode(epyqlib.canneo.Frame, TreeNode):
    def __init__(self, message=None, tx=False, frame=None,
                 multiplex_value=None, signal_class=SignalNode,
                 mux_frame=None, parent=None, **kwargs):
        epyqlib.canneo.Frame.__init__(self, frame=frame,
                                   multiplex_value=multiplex_value,
                                   signal_class=signal_class,
                                   mux_frame=mux_frame,
                                   parent=parent,
                                   **kwargs)
        TreeNode.__init__(self, parent)

        self.fields = Columns()
        self.last_time = None

        self.tx = tx
        self._send_checked = False

        self.count = {
            'tx': 0,
            'rx': 0,
        }

        for signal in self.signals:
            signal.tx = self.tx
            self.append_child(signal)

        identifier = epyqlib.canneo.format_identifier(self.id, self.extended)

        if self.mux_name is None:
            name = self.name
        else:
            name = '{} - {}'.format(self.name, self.mux_name)

        count = self.count['tx'] if self.tx else self.count['rx']

        self.fields = Columns(id=identifier,
                              name=name,
                              length='{} B'.format(self.size),
                              value='-',
                              dt='-',
                              count=str(count))

    @property
    def send_checked(self):
        return self._send_checked

    @send_checked.setter
    def send_checked(self, value):
        old = self._send_checked

        # TODO: move this validation to dt to check itself
        if not isinstance(self.dt, float):
            if value != Qt.Unchecked:
                self.dt = 0.1

        self._send_checked = value

        if self._send_checked != old:
            self.update_timer()

    def checked(self, column):
        if column == Columns.indexes.dt:
            return self.send_checked

        return Qt.Unchecked

    @property
    def dt(self):
        return self.fields.dt

    @dt.setter
    def dt(self, value):
        old = self.fields.dt

        if value == '':
            self.fields.dt = value
            self.send_checked = Qt.Unchecked
        else:
            # TODO: move this validation to dt to check itself
            check_it = not isinstance(self.fields.dt, float)
            self.fields.dt = float(value)
            if check_it:
                self.send_checked = Qt.Checked

        if self.fields.dt != old:
            self.update_timer()

    def update_timer(self):
        if self.send_checked == Qt.Unchecked:
            self.cyclic_request(self, None)
        else:
            self.cyclic_request(self, self.dt)

    def extract_message(self, message, verify=True):
        # TODO: stop calling this from txrx and use the standard Frame reception

        if verify:
            # TODO: make sure the message matches the id/type and length
            pass

        # TODO: I think this is not needed
        # self.message = message

        self.fields.id = epyqlib.canneo.format_identifier(
                message.arbitration_id, message.id_type)

        # self.fields.name = self.name

        self.fields.length = '{} B'.format(message.dlc)
        self.fields.value = epyqlib.canneo.format_data(tuple(message.data))
        # if self.last_time == message.timestamp:
        #     raise Exception('message already received {message}'
        #                     .format(**locals()))
        if self.last_time is None:
            self.fields.dt = '-'
        else:
            self.fields.dt = '{:.4f}'.format(message.timestamp - self.last_time)
        self.last_time = message.timestamp

        self.count['rx'] += 1

        if not self.tx:
            self.fields.count = str(self.count['rx'])

        epyqlib.canneo.Frame.message_received(self, message)

    def _sent(self):
        super()._sent()

        self.count['tx'] += 1

        if self.tx:
            self.fields.count = str(self.count['tx'])
            self.tree_parent.changed.emit(
                self, Columns.indexes.count,
                self, Columns.indexes.count,
                [Qt.DisplayRole])

    def message_received(self, msg):
        super().message_received(msg=msg)

    def unique(self):
        return id(self)

    def set_data(self, data):
        self.fields.value = data
        self.update()
        self.frame.update_from_signals()

    def update_from_signals(self):
        epyqlib.canneo.Frame.update_from_signals(self)
        self.fields.value = epyqlib.canneo.format_data(self.data)
        # TODO: send should update, not the other way around like it is
        self._send()


class TxRx(TreeNode, epyqlib.canneo.QtCanListener):
    # TODO: just Rx?
    changed = epyqlib.utils.qt.Signal(
        TreeNode,
        int,
        TreeNode,
        int,
        list,
    )
    begin_insert_rows = epyqlib.utils.qt.Signal(
        TreeNode,
        int,
        int,
    )
    end_insert_rows = epyqlib.utils.qt.Signal()

    def __init__(self, tx, neo=None, bus=None, parent=None):
        TreeNode.__init__(self)
        epyqlib.canneo.QtCanListener.__init__(self, parent=parent)

        self.bus = bus

        self.tx = tx
        self.rx = not self.tx
        self.neo = neo
        self.messages = {}

        if self.rx:
            self.message_received_signal.connect(self.message_received)

        if self.tx:
            for frame in self.neo.frames:
                self.add_message_node(node=frame)

        if self.bus is not None:
            self.remapper = epyqlib.canneo.QtCanListener(
                receiver=self.message_sent)
            self.bus.tx_notifier.add(self.remapper)
            self.sent_messages = {}

        # TODO: this should probably be done in the view but this is easier for now
        #       Tx can't be added to later (yet)
        #       Rx will add in order received
        self.children.sort(key=lambda c: c.name)

        self.fields = Columns(id='',
                      name='',
                      length='',
                      value='',
                      dt='',
                      count='')

        self.update_timer = QTimer()
        # 81 seems to provide nice variation in the count on 10ms messages
        # but is still a bit of a 'random' selection.
        # TODO: pick a number based on the cycle times in the .sym
        self.update_timer.setInterval(81)
        self.update_timer.timeout.connect(self.gui_update)
        self.update_timer.start()

    def gui_update(self):
        # TODO: this is hacky but without it the tx widget would get values
        #       overwritten while they were being edited keeping the user
        #       from actually being able to change them
        if self.rx:
            self.changed.emit(
                self, 0,
                self, 1,
                [Qt.DisplayRole])

    def set_node_id(self, node_id):
        # TODO: I think this can go away
        self.node_id = node_id

    def add_message(self, message=can.Message(), id=None, tx=False):
        message_node = self.neo.get_multiplex(message)[0]

        if id is None:
            id = self.generate_id(message=message)

        if message_node is None:
            # TODO: yuck, stop using the matrix
            frame = canmatrix.canmatrix.Frame(
                id=message.arbitration_id,
                name='',
                size=message.dlc,
            )
            message_node = MessageNode(message=message, tx=tx, frame=frame)
            self.neo.frames = self.neo.frames + (message_node,)

        message_node.send.connect(self.send)
        self.messages[id] = message_node

        index = len(self.children)
        # TODO: move to TreeNode?
        self.begin_insert_rows.emit(self, index, index)
        self.append_child(message_node)
        self.end_insert_rows.emit()

    def add_message_node(self, node):
        node.send.connect(self.send)

        index = len(self.children)
        # TODO: move to TreeNode?
        self.begin_insert_rows.emit(self, index, index)
        self.append_child(node)
        self.end_insert_rows.emit()

    def generate_id(self, message):
        multiplex_value = self.neo.get_multiplex(message)[1]

        return (message.arbitration_id,
                message.id_type,
                multiplex_value)

    def message_received(self, msg):
        id = self.generate_id(message=msg)

        try:
            message = self.messages[id]
        except KeyError:
            self.add_message(message=msg,
                             id=id)
            message = self.messages[id]

        message.extract_message(msg)

        # for column in [Columns.indexes.value, Columns.indexes.dt,
        #                Columns.indexes.count]:
        #     self.changed.emit(message, column,
        #                       message, column,
        #                       [Qt.DisplayRole])
        #
        # for child in message.children:
        #     self.changed.emit(
        #         child, Columns.indexes.value,
        #         child, Columns.indexes.value,
        #         [Qt.DisplayRole])

    def message_sent(self, message):
        id = self.generate_id(message=message)

        node = self.sent_messages.get(id)

        if node is None:
            for node in self.children:
                if (node.id, node.extended, node.mux_value) == id:
                    self.sent_messages[id] = node
                    break
            else:
                # not recognized
                self.sent_messages[id] = None
                return

        if node is not None:
            node._sent()

    def unique(self):
        # TODO: actually identify the object
        return '-'

    def send(self, message, on_success=None):
        self.bus.send(message, on_success=on_success)

    def __str__(self):
        return 'Indexes: \n' + '\n'.join([str(i) for i in self.children])


class Columns(AbstractColumns):
    _members = ['id', 'length', 'name', 'value', 'dt', 'count']

Columns.indexes = Columns.indexes()


class TxRxModel(epyqlib.pyqabstractitemmodel.PyQAbstractItemModel):
    def __init__(self, root, parent=None):
        checkbox_columns = Columns.fill(False)
        if root.tx:
            checkbox_columns.dt = True

        epyqlib.pyqabstractitemmodel.PyQAbstractItemModel.__init__(
                self, root=root, checkbox_columns=checkbox_columns,
                parent=parent)

        self.headers = Columns(id='ID',
                               length='Length',
                               name='Name',
                               value='Value',
                               dt='Cycle Time',
                               count='Count')

    def flags(self, index):
        flags = epyqlib.pyqabstractitemmodel.PyQAbstractItemModel.flags(self, index)

        node = self.node_from_index(index)
        if node.tx:
            if index.column() == Columns.indexes.value:
                if isinstance(node, SignalNode):
                    allow = node.multiplex is not True
                else:
                    allow = False

                if allow:
                    flags |= Qt.ItemIsEditable

            if index.column() == Columns.indexes.dt:
                if isinstance(node, MessageNode):
                    mask = Qt.ItemIsEditable
                    mask |= Qt.ItemIsUserCheckable
                    if node.user_send_control:
                        flags |= mask
                    else:
                        flags &= ~mask

        return flags

    def setData(self, index, data, role=None):
        if index.column() == Columns.indexes.value:
            if role == Qt.EditRole:
                node = self.node_from_index(index)
                try:
                    node.set_data(data)
                except ValueError:
                    return False
                else:
                    self.dataChanged.emit(index, index)
                    return True

            return False

        if index.column() == Columns.indexes.dt:
            if role == Qt.EditRole:
                node = self.node_from_index(index)
                try:
                    node.dt = data
                except ValueError:
                    return False

                self.dataChanged.emit(index, index)
                return True
            if role == Qt.CheckStateRole:
                node = self.node_from_index(index)
                node.send_checked = data
                return True

        return False



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

import functools
import textwrap
import time

from PyQt5 import QtWidgets

import epyqlib.parameteredit_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class ParameterEdit(QtWidgets.QWidget):
    def __init__(self, parent=None, in_designer=False, edit=None, nv=None,
                 dialog=None, esc_action=None):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self._edit = edit
        self._dialog = dialog
        self.nv = nv

        self.ui = epyqlib.parameteredit_ui.Ui_Form()
        self.ui.setupUi(self)

        self.ui.to_device.set_signal(self.nv)
        self.nv.status_signal.value_changed.connect(self.nv.value_changed)

        self.ui.save_to_nv_button.clicked.connect(self.save_to_nv)

        self.ui.to_device.edited.connect(self.edited)

        self.ui.description.setText(self.ui.to_device.toolTip())

        self.esc_action = esc_action
        self.ui.esc_button.clicked.connect(self.esc)

    def esc(self, checked):
        if self.esc_action is not None:
            self.esc_action()

    def edited(self, value):
        self.nv.read_from_device()
        time.sleep(0.05)
        self.nv.set_human_value(value)
        self.nv.write_to_device()

    def save_to_nv(self):
        focus_self = functools.partial(
            self.parent().setCurrentWidget,
            self
        )

        # TODO: CAMPid 93849811216123127753953680713426
        def inverter_to_nv():
            self.nv.tree_parent.module_to_nv()
            focus_self()

        self._dialog.focus(ok_action=inverter_to_nv,
                         cancel_action=focus_self,
                         label=textwrap.dedent('''\
                             Save all parameters to NV?''')
                         )

if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



import collections
import contextlib
import decimal
import functools
import inspect
import itertools
import json
import locale
import logging
import string
import sys
import uuid
import weakref
from natsort import natsorted

import attr
import graham
import graham.fields
import marshmallow
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets
import PyQt5.QtCore

import epyqlib.abstractcolumns
import epyqlib.delegates
import epyqlib.treenode
import epyqlib.utils.general
import epyqlib.utils.qt

# See file COPYING in this source tree
__copyright__ = 'Copyright 2017, EPC Power Corp.'
__license__ = 'GPLv2+'


logger = logging.getLogger()


class NotFoundError(Exception):
    pass


class MultipleFoundError(Exception):
    pass


def create_str_attribute(default=''):
    return attr.ib(
        default=default,
        converter=str,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )


def create_str_or_none_attribute(default=None):
    return attr.ib(
        default=default,
        converter=to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )


def create_name_attribute(default=None):
    return attr.ib(
        default=default,
        converter=to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )


def create_reference_attribute():
    attribute = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=epyqlib.attrsmodel.Reference(allow_none=True),
        ),
    )
    epyqlib.attrsmodel.attrib(
        attribute=attribute,
        no_column=True,
    )

    return attribute


def create_checkbox_attribute(default=False):
    return attr.ib(
        default=default,
        converter=epyqlib.attrsmodel.two_state_checkbox,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(),
        ),
    )


def create_integer_attribute(default=0):
    return attr.ib(
        default=default,
        converter=int,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(),
        ),
    )


class NotAllowedCharacterError(Exception):
    @classmethod
    def build(cls, value, allowed):
        not_allowed = set(value) - set(allowed)

        return cls(
            'Characters {!r} from {!r} are not in allowed set {!r}'.format(
                ''.join(sorted(not_allowed)),
                value,
                ''.join(sorted(allowed)),
            )
        )


class NotAllowedFirstCharacterError(Exception):
    @classmethod
    def build(cls, first_character, allowed):
        return cls(
            'First character {!r} not allowed, must be one of {!r}'.format(
                first_character,
                ''.join(sorted(allowed)),
            )
        )


@attr.s
class LimitedStringConverter:
    allowed = attr.ib()
    allowed_first_character = attr.ib()

    @classmethod
    def build(cls, allowed, not_allowed_first):
        allowed = set(allowed)
        not_allowed_first = set(not_allowed_first)

        return cls(
            allowed=allowed,
            allowed_first_character=allowed - not_allowed_first,
        )

    def __call__(self, value):
        value = str(value)
        value_set = set(value)

        first_character = value[0]

        if first_character not in self.allowed_first_character:
            raise NotAllowedFirstCharacterError.build(
                first_character=first_character,
                allowed=self.allowed_first_character,
            )

        if not value_set.issubset(self.allowed):
            raise NotAllowedCharacterError.build(
                value=value,
                allowed=self.allowed
            )

        return value

    def suggest(self, value):
        suggestion = ''.join(
            character
            for character in value
            if character in self.allowed
        )

        if suggestion[:1] not in self.allowed_first_character:
            if '_' in self.allowed_first_character:
                suggestion = '_' + suggestion
            else:
                while suggestion[:1] not in self.allowed_first_character:
                    suggestion = suggestion[1:]
                    if len(suggestion) == 0:
                        break

        return suggestion


def create_limited_string_attribute(default, allowed, not_allowed_first):
    return attr.ib(
        default=default,
        converter=LimitedStringConverter.build(
            allowed=allowed,
            not_allowed_first=not_allowed_first,
        ),
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )


def create_code_identifier_string_attribute(default):
    return create_limited_string_attribute(
        default=default,
        allowed=string.ascii_letters + string.digits + '_',
        not_allowed_first=string.digits,
    )


# TODO: CAMPid 8695426542167924656654271657917491654
def name_from_uuid(node, value, model):
    if value is None:
        return None

    try:
        target_node = model.node_from_uuid(value)
    except NotFoundError:
        return str(value)

    return target_node.name


def names_from_uuid_list(node, value, model):
    if value is None:
        return None

    target_nodes = model.nodes_from_uuid_list(value)

    names = []
    for n in target_nodes:
        names.append(n.name)
    return names


@attr.s
class Column:
    name = attr.ib()
    fields = attr.ib()

    def __iter__(self):
        return iter(self.fields)

    def items(self):
        return self.fields.items()


class ConsistencyError(Exception):
    pass


@attr.s
class Metadata:
    name = attr.ib(default=None)
    data_display = attr.ib(default=None)
    editable = attr.ib(default=True)
    human_name = attr.ib(default=None)
    converter = attr.ib(default=None)
    no_column = attr.ib(default=False)
    delegate = attr.ib(default=None)
    updating = attr.ib(default=False)


metadata_key = object()


@attr.s
class Attributes:
    fields = attr.ib()


def ify():
    def inner(cls):
        class Fields:
            def __iter__(self):
                return (
                    getattr(self, field.name)
                    for field in attr.fields(type(self))
                )

        for field in attr.fields(cls):
            setattr(Fields, field.name, attr.ib())

        Fields = attr.s(Fields)

        field_metadata = collections.defaultdict(dict)
        for field in attr.fields(cls):
            metadata = field.metadata.get(metadata_key)

            extras = {}
            if field.name == 'children':
                extras['no_column'] = True

            if metadata is None:
                metadata = Metadata(name=field.name, **extras)
            else:
                metadata = attr.evolve(metadata, name=field.name, **extras)

            if metadata.converter is None:
                metadata.converter = field.converter

            field_metadata[field.name] = metadata

        setattr(
            cls,
            attribute_name,
            Attributes(
                fields=Fields(**field_metadata),
            ),
        )

        return cls

    return inner


attribute_name = epyqlib.utils.general.identifier_path(Attributes)


def attributes(cls):
    return getattr(cls, attribute_name)


def fields(cls):
    return attributes(cls).fields

def list_selection_roots(cls):
    d = {}
    for field in fields(cls):
        if field.delegate is not None:
            if field.delegate.list_selection_root:
                d[field.name] = field.delegate.list_selection_root
    return d


def attrib(*args, attribute, **kwargs):
    # https://github.com/python-attrs/attrs/issues/278
    if len(attribute.metadata) == 0:
        attribute.metadata = {}

    attribute.metadata[metadata_key] = Metadata(*args, **kwargs)

    return attribute


def data_processor(cls, data_field, attribute_field):
    metadata = getattr(attributes(cls).fields, data_field.name)
    d = getattr(metadata, attribute_field.name)

    return d


@attr.s
class Columns:
    columns = attr.ib()

    def __iter__(self):
        return iter(self.columns)

    def __getitem__(self, item):
        if isinstance(item, str):
            column, = (
                column
                for column in self.columns
                if column.name == item
            )
            return column

        if isinstance(item, tuple):
            column, = (
                column
                for column in self.columns
                if column.fields[item[0]] == item[1]
            )
            return column

        return self.columns[item]

    def index_of(self, item):
        if isinstance(item, str):
            index, = (
                index
                for index, column in enumerate(self.columns)
                if column.name == item
            )
        elif isinstance(item, tuple):
            index, = (
                index
                for index, column in enumerate(self.columns)
                if column.fields[item[0]] == item[1]
            )

        return index


def columns(*columns):
    def _name(column):
        cls, field_name = column

        field = getattr(fields(cls), field_name)
        name = field.human_name

        if name is None:
            name = field_name.replace('_', ' ').title()

        return name

    return Columns(
        columns=tuple(
            Column(name=_name(c[0]), fields=dict(c))
            for c in columns
        ),
    )


@attr.s
class Types:
    types = attr.ib(
        converter=(
            lambda types:
            collections.OrderedDict((t.__name__, t) for t in types)
        ),
        default=(),
    )
    external_list_selection_roots = attr.ib(factory=set)

    def __attrs_post_init__(self):
        for t in self.types.values():
            add_addable_types(cls=t, types=self)

    def __getitem__(self, item):
        return tuple(self.types.values())[item]

    def resolve(self, type_, default=None):
        if type_ is None and default is not None:
            return default

        if isinstance(type_, str):
            return self.types[type_]

        if isinstance(type_, marshmallow.Schema):
            return type_.data_class

        return type_
    
    def list_selection_roots(self):
        roots = set(self.external_list_selection_roots)
        for v in self.types.values():
            t = list_selection_roots(v)
            roots.update(t.values())

        return roots


def create_addable_types(types):
    return collections.OrderedDict((
        (
            type_.__name__,
            type_,
        )
        for type_ in types
    ))


def add_addable_types(cls, attribute_name='children', types=None):
    if types is None:
        types = Types()

    if hasattr(cls, 'addable_types') or hasattr(cls, 'all_addable_types'):
        return

    @classmethod
    def addable_types(cls):
        if cls.addable_types_cache is None:
            field = graham.schema(cls).fields.get(attribute_name)
            if field is None:
                return {}

            resolved_types = tuple(
                types.resolve(type_=t.nested, default=cls)
                for t in field.instances
            )

            cls.addable_types_cache = create_addable_types(resolved_types)

        return cls.addable_types_cache

    cls.addable_types = addable_types
    cls.all_addable_types = addable_types
    cls.addable_types_cache = None
    cls.addable_types()

    return cls


def check_just_children(self, models):
    # TODO: ugh, circular dependencies
    import epyqlib.checkresultmodel

    if len(self.children) == 0:
        return None

    child_results = [
        child.check(models=models)
        for child in self.children
    ]
    child_results = [
        results
        for results in child_results
        if results is not None
    ]

    if len(child_results) == 0:
        return None

    return epyqlib.checkresultmodel.Node.build(
        name=getattr(self, 'name', ''),
        node=self,
        child_results=child_results,
    )


def check_children(f):
    def wrapper(self, models):
        # TODO: ugh, circular dependencies
        import epyqlib.checkresultmodel

        result = check_just_children(self, models=models)

        if result is None:
            result = epyqlib.checkresultmodel.Node.build(
                name=getattr(self, 'name', ''),
                node=self,
            )

        result = f(self=self, result=result, models=models)

        if len(result.children) == 0:
            return None

        return result

    return wrapper


def Root(default_name, valid_types):
    @graham.schemify(tag='root')
    @ify()
    @epyqlib.utils.qt.pyqtify()
    @attr.s(hash=False)
    class Root(epyqlib.treenode.TreeNode):
        name = attr.ib(
            default=default_name,
        )
        graham.attrib(
            attribute=name,
            field=marshmallow.fields.String(),
        )

        children = attr.ib(
            default=attr.Factory(list),
        )
        graham.attrib(
            attribute=children,
            field=graham.fields.MixedList(fields=(
                marshmallow.fields.Nested(graham.schema(type_))
                for type_ in valid_types
                # marshmallow.fields.Nested('Group'),
                # marshmallow.fields.Nested(graham.schema(Leaf)),
            )),
        )

        model = attr.ib(default=None)

        uuid = attr_uuid()

        def __attrs_post_init__(self):
            super().__init__()

        def can_drop_on(self, node):
            return isinstance(node, tuple(self.addable_types().values()))

        remove_old_on_drop = default_remove_old_on_drop
        child_from = default_child_from
        internal_move = default_internal_move

        def check_and_append(self, models, parent=None):
            # TODO: ugh, circular dependencies
            import epyqlib.checkresultmodel

            if parent is None:
                parent = epyqlib.checkresultmodel.Root()

            for child in self.children:
                result = child.check(models=models)

                if result is not None:
                    parent.append_child(result)

            return parent

        def check(self, models):
            return self.check_and_append(models=models)

        @staticmethod
        def can_delete(node=None):
            if node is None:
                return False

            return True

    return Root


def convert_uuid(x):
    if x == '':
        return None

    if x is None or isinstance(x, uuid.UUID):
        return x

    return uuid.UUID(x)


def convert_uuid_list(x):
    if x is None:
        return None

    l = []
    for y in x:
        if isinstance(y, uuid.UUID):
            l.append(y)
        else:
            l.append(convert_uuid(y.uuid))

    return l


def attr_uuid_list(
        metadata=None,
        human_name='UUID List',
        data_display=None,
        list_selection_root=None,
        no_graham=False,
        default=attr.Factory(list),
        **field_options,
):
    if metadata is None:
        metadata = {}

    attribute = attr.ib(
        default=default,
        converter=convert_uuid_list,
        metadata=metadata,
    )
    if not no_graham:
        graham.attrib(
            attribute=attribute,
            field=marshmallow.fields.List(marshmallow.fields.UUID(**field_options), **field_options),
        )
    attrib(
        attribute=attribute,
        human_name=human_name,
        data_display=data_display,
        delegate=epyqlib.attrsmodel.RootDelegateCache(
            list_selection_root=list_selection_root,
            multi_select=True,
        ),
    )

    return attribute


def attr_uuid(
        metadata=None,
        human_name='UUID',
        data_display=None,
        list_selection_root=None,
        list_selection_path=None,
        override_delegate=None,
        no_graham=False,
        default=attr.Factory(uuid.uuid4),
        editable=True,
        no_column=False,
        **field_options,
):
    if metadata is None:
        metadata = {}

    attribute = attr.ib(
        default=default,
        converter=convert_uuid,
        metadata=metadata,
    )
    if not no_graham:
        graham.attrib(
            attribute=attribute,
            field=marshmallow.fields.UUID(**field_options),
        )

    if list_selection_path is not None:
        if list_selection_root is not None:
            raise MultipleFoundError(
                    'list_selection_path and list_selection_root both definded'
            )
        else:
            attrib(
                attribute=attribute,
                human_name=human_name,
                data_display=data_display,
                delegate=CustomDelegate(
                    list_selection_path=list_selection_path,
                    override_delegate=override_delegate,
                ),
                editable=editable,
                no_column=no_column,
            )
    else:
        attrib(
            attribute=attribute,
            human_name=human_name,
            data_display=data_display,
            delegate=RootDelegateCache(
                list_selection_root=list_selection_root
            ),
            editable=editable,
            no_column=no_column,
        )

    return attribute


def to_decimal_or_none(s):
    if s is None:
        return None

    if isinstance(s, str) and len(s) == 0:
        return None

    if isinstance(s, str):
        s = locale.delocalize(s)

    try:
        result = decimal.Decimal(s)
    except decimal.InvalidOperation as e:
        raise ValueError('Invalid number: {}'.format(repr(s))) from e

    return result


def to_str_or_none(s):
    if s is None:
        return None

    if isinstance(s, str):
        if len(s) == 0:
            return None

        return s

    return str(s)


def to_int_or_none(s):
    if s is None:
        return None

    if isinstance(s, str) and len(s) == 0:
        return None

    if isinstance(s, str):
        s = locale.delocalize(s)

    try:
        result = int(s)
    except ValueError as e:
        raise ValueError('Invalid number: {}'.format(repr(s))) from e

    return result


def two_state_checkbox(v):
    return v in (QtCore.Qt.Checked, True)


def check_uuids(*roots):
    def collect(node, uuids):
        if node.uuid is not None:
            if node.uuid in uuids:
                raise Exception('Duplicate uuid found: {}'.format(node.uuid))

            uuids.add(node.uuid)

    def set_nones(node, uuids):
        if node.uuid is None:
            while node.uuid is None:
                u = uuid.uuid4()
                if u not in uuids:
                    node.uuid = u
                    uuids.add(node.uuid)

    uuids = set()

    for root in set(roots):
        root.traverse(call_this=collect, payload=uuids, internal_nodes=True)

    for root in set(roots):
        root.traverse(call_this=set_nones, payload=uuids, internal_nodes=True)


def childless_can_delete(self, node=None):
    if node is not None:
        raise ConsistencyError(
            'No children to be considered'
        )

    return self.tree_parent.can_delete(node=self)


def default_remove_old_on_drop(self, node):
    return node.find_root() == self.find_root()


@staticmethod
def default_child_from(node):
    return node


def default_internal_move(self, node, node_to_insert_before):
    return False


@classmethod
def empty_all_addable_types(cls):
    return epyqlib.attrsmodel.create_addable_types(())


def empty_addable_types(self):
    return {}


def to_source_model(index):
    model = index.model()
    while not isinstance(model, QtGui.QStandardItemModel):
        index = model.mapToSource(index)
        model = index.model()

    return index


def create_delegate(parent=None):
    selector = DelegateSelector(parent=parent)
    delegate = epyqlib.delegates.Dispatch(
        selector=selector.select,
        parent=parent,
    )

    return delegate


def get_connection_id(parent, child):
    if parent is not None:
        parent = parent.uuid

    return (parent, child.uuid)


def hide_popup(self):
    QtCore.QCoreApplication.postEvent(
        self,
        QtGui.QKeyEvent(
            QtCore.QEvent.KeyPress,
            QtCore.Qt.Key_Enter,
            QtCore.Qt.NoModifier,
        ),
    )


@attr.s
class RootDelegateCache:
    list_selection_root = attr.ib()
    text_column_name = attr.ib(default='Name')
    cached_delegate = attr.ib(default=None)
    multi_select = attr.ib(default=False)

    def get_delegate(self, model, parent):
        if self.cached_delegate is not None:
            return self.cached_delegate

        root_node = model.list_selection_roots[
            self.list_selection_root
        ]

        delegate = EnumerationDelegate
        if self.multi_select:
            delegate = EnumerationDelegateMulti

        self.cached_delegate = delegate(
            text_column_name=self.text_column_name,
            root=root_node,
            parent=parent,
        )

        return self.cached_delegate


@attr.s
class CustomDelegate:
    list_selection_path = attr.ib(default=None)
    override_delegate = attr.ib(default=None)
    text_column_name = attr.ib(default='Name')

    def get_delegate(self, node, parent):
        root_node = node
        for element in self.list_selection_path:
            if element == '/':
                root_node = root_node.find_root()
            elif element == '..':
                root_node = root_node.tree_parent
            else:
                root_node = root_node.child_by_name(element)

        delegate = EnumerationDelegate
        if self.override_delegate is not None:
            delegate = self.override_delegate

        return delegate(
            text_column_name=self.text_column_name,
            root=root_node,
            parent=parent,
        )


class DelegateSelector:
    def __init__(self, parent=None):
        self.parent = parent
        self.regular = QtWidgets.QStyledItemDelegate(parent)
        self.enumerations = {}


    def select(self, index):
        index = to_source_model(index)
        model = index.model()
        item = model.itemFromIndex(index)
        node = item.data(epyqlib.utils.qt.UserRoles.node)
        field_name = item.data(epyqlib.utils.qt.UserRoles.field_name)
        model = item.data(epyqlib.utils.qt.UserRoles.attrs_model)

        metadata = getattr(fields(type(node)), field_name)
        
        if metadata.delegate is not None:
            node_or_model = model
            if isinstance(metadata.delegate, epyqlib.attrsmodel.CustomDelegate):
                node_or_model = node
            delegate = metadata.delegate.get_delegate(node_or_model, self.parent)
        else:
            delegate = self.regular

        return delegate


# TODO: CAMPid 374895478431714307074310
class CustomCombo(PyQt5.QtWidgets.QComboBox):
    def hidePopup(self):
        super().hidePopup()
        hide_popup(self)


class CustomMulti(PyQt5.QtWidgets.QListWidget):
    def hidePopup(self):
        super().hidePopup()
        hide_popup(self)


# TODO: CAMPid 05470876451650542168542
class EnumerationDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, text_column_name, root, parent):
        super().__init__(parent)

        self.text_column_name = text_column_name
        self.root = root

    def createEditor(self, parent, option, index):
        return CustomMulti(parent=parent)

    def setEditorData(self, editor, index):
        super().setEditorData(editor, index)

        model_index = to_source_model(index)
        model = model_index.model()

        raw = model.data(model_index, epyqlib.utils.qt.UserRoles.raw)
        editor.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        for node in self.root.children:
            it = PyQt5.QtWidgets.QListWidgetItem(editor)
            it.setText(node.name)
            it.setData(epyqlib.utils.qt.UserRoles.raw, node.uuid)
            if raw == node.uuid:
                it.setSelected(True)

        editor.setMinimumSize(editor.sizeHint())
        editor.show()

    def setModelData(self, editor, model, index):
        index = epyqlib.utils.qt.resolve_index_to_model(index)
        model = index.model()

        selected_items = editor.selectedItems()

        if len(selected_items) == 0:
            selected_uuid = ''
        else:
            selected_item, = selected_items
            selected_uuid = selected_item.data(epyqlib.utils.qt.UserRoles.raw)
            selected_uuid = str(selected_uuid)

        model.setData(index, selected_uuid)


# TODO: CAMPid 05470876451650542168542
class EnumerationDelegateMulti(QtWidgets.QStyledItemDelegate):
    def __init__(self, text_column_name, root, parent):
        super().__init__(parent)

        self.text_column_name = text_column_name
        self.root = root

    def createEditor(self, parent, option, index):
        return CustomMulti(parent=parent)

    def setEditorData(self, editor, index):
        super().setEditorData(editor, index)

        model_index = to_source_model(index)
        model = model_index.model()

        raw = model.data(model_index, epyqlib.utils.qt.UserRoles.raw)
        editor.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        for node in self.root.children:
            it = PyQt5.QtWidgets.QListWidgetItem(editor)
            it.setText(node.name)
            it.uuid = node.uuid
            if raw is not None:
                for r in raw:
                    if r == it.uuid:
                        it.setSelected(True)

        editor.setMinimumSize(editor.sizeHint())
        editor.show()

    def setModelData(self, editor, model, index):
        index = epyqlib.utils.qt.resolve_index_to_model(index)
        model = index.model()

        selected_items = natsorted(editor.selectedItems())
        model.setData(index, selected_items)


class PyQStandardItemModel(QtGui.QStandardItemModel):
    # maybe events instead?
    # http://doc.qt.io/qt-5/dnd.html

    def canDropMimeData(self, *args, **kwargs):
        return self.can_drop_mime_data(*args, **kwargs)

    def mimeTypes(self, *args, **kwargs):
        return self.mime_types(*args, **kwargs)

    def mimeData(self, *args, **kwargs):
        return self.mime_data(*args, **kwargs)

    def dropMimeData(self, *args, **kwargs):
        return self.drop_mime_data(*args, **kwargs)

    def supportedDropActions(self, *args, **kwargs):
        return self.supported_drop_actions(*args, **kwargs)

    @classmethod
    def build(
            cls,
            *args,
            can_drop_mime_data,
            mime_types,
            mime_data,
            drop_mime_data,
            supported_drop_actions,
            **kwargs,
    ):
        model = cls(*args, **kwargs)
        model.can_drop_mime_data = can_drop_mime_data
        model.mime_types = mime_types
        model.mime_data = mime_data
        model.drop_mime_data = drop_mime_data
        model.supported_drop_actions = supported_drop_actions

        return model


class Model:
    def __init__(self, root, columns, drop_sources=(), parent=None):
        self.root = root
        self.root.model = self
        self._all_items_dict = {}
        self.node_to_item = {}
        self.uuid_to_node = {}

        self.model = PyQStandardItemModel.build(
            can_drop_mime_data=self.canDropMimeData,
            mime_types=self.mimeTypes,
            mime_data=self.mimeData,
            drop_mime_data=self.dropMimeData,
            supported_drop_actions=self.supportedDropActions,
        )

        self.model.invisibleRootItem().setData(
            self,
            epyqlib.utils.qt.UserRoles.attrs_model,
        )

        self.mime_type = 'application/com.epcpower.pm.attrsmodel'

        self.columns = columns
        self.header_items = [
            QtGui.QStandardItem(column.name)
            for column in self.columns
        ]

        for i, item in enumerate(self.header_items):
            self.model.setHorizontalHeaderItem(i, item)

        self.droppable_from = set()

        self.connected_signals = {}

        self.list_selection_roots = {}

        self.add_drop_sources(*drop_sources)

        check_uuids(self.root)

        self.pyqtify_connect(None, self.root)
        self.model.itemChanged.connect(self.item_changed)

    def add_drop_sources(self, *sources):
        self.droppable_from.update(sources)
        roots = [
            model.root
            for model in {self} | self.droppable_from
        ]

        check_uuids(*roots)

    def get_field(self, index):
        c = index.column()
        t = type(self.node_from_index(index))
        name = self.columns[c].fields.get(t)

        if name is None:
            return None

        return getattr(attributes(t).fields, name)

    def item_changed(self, item):
        node = self.node_from_item(item)

        field_name = item.data(
            epyqlib.utils.qt.UserRoles.field_name,
        )
        field_metadata = getattr(
            attributes(node).fields,
            field_name,
        )
        if field_metadata.updating:
            return

        if field_name is None:
            # TODO: why is it getting changed if there's nothing there?
            return
        datum = item.data(QtCore.Qt.DisplayRole)
        if field_metadata.converter is not None:
            if field_metadata.converter == two_state_checkbox:
                datum = item.data(QtCore.Qt.CheckStateRole)

            try:
                datum = field_metadata.converter(datum)
            except:
                item.setData(getattr(node, field_name), QtCore.Qt.DisplayRole)
                raise

        setattr(node, field_name, datum)

    def pyqtify_connect(self, parent, child):
        def visit(node, nodes):
            if node is child:
                this_parent = parent
            else:
                this_parent = node.tree_parent

            nodes.append({'parent': this_parent, 'child': node})

        nodes = []
        child.traverse(call_this=visit, payload=nodes, internal_nodes=True)

        self.uuid_to_node.update({
            child.uuid: child
            for child in (
                d['child']
                for d in nodes
            )
        })

        for kwargs in nodes:
            self._pyqtify_connect(**kwargs)

    def item_from_node(self, node):
        if node is self.root:
            return self.model.invisibleRootItem()

        return self.node_to_item[node]

    def node_from_item(self, item):
        return item.data(epyqlib.utils.qt.UserRoles.node)

    def _pyqtify_connect(self, parent, child):
        connections = {}
        connection_id = get_connection_id(parent=parent, child=child)
        if connection_id in self.connected_signals:
            raise ConsistencyError('already connected: {}'.format((parent.uuid, child.uuid)))
        self.connected_signals[connection_id] = connections

        if child is self.root:
            root_item = self.model.invisibleRootItem()
            root_item.setData(child, epyqlib.utils.qt.UserRoles.node)
            root_item.setText('root')
        else:
            if parent is self.root:
                parent_item = self.model.invisibleRootItem()
            else:
                parent_item = self.item_from_node(parent)
            row = parent.row_of_child(child)

            items = []

            changed_signals = epyqlib.utils.qt.pyqtified(child).changed

            uneditable_highlight = QtGui.QColor('grey')
            uneditable_highlight.setAlphaF(0.4)
            droppable_highlight = QtGui.QColor('orange')
            droppable_highlight.setAlphaF(0.4)
            droppable_row = hasattr(row, 'addable_types') or hasattr(row, 'all_addable_types')

            for i, column in enumerate(self.columns):
                field_name = column.fields.get(type(child))

                item = QtGui.QStandardItem()
                editable = False
                has_field = False
                field_for_column = column.fields.get(type(child))
                if field_for_column is not None:
                    metadata = getattr(fields(type(child)), field_for_column)
                    editable = metadata.editable
                    has_field = True

                item.setEditable(editable)
                if not editable and has_field:
                    if droppable_row:
                        item.setData(
                            uneditable_highlight,
                            QtCore.Qt.ItemDataRole.BackgroundRole,
                        )
                    else:
                        item.setData(
                            droppable_highlight,
                            QtCore.Qt.ItemDataRole.BackgroundRole,
                        )

                if i == 0:
                    self._all_items_dict[(
                        item.data(epyqlib.utils.qt.UserRoles.node),
                        item.column(),
                    )] = item
                    self.node_to_item[child] = item
                item.setData(child, epyqlib.utils.qt.UserRoles.node)
                item.setData(i, epyqlib.utils.qt.UserRoles.column_index)
                item.setData(
                    field_name,
                    epyqlib.utils.qt.UserRoles.field_name,
                )
                item.setData(
                    self,
                    epyqlib.utils.qt.UserRoles.attrs_model,
                )

                if field_name is not None:
                    checkable = (
                        getattr(fields(type(child)), field_name).converter
                        == two_state_checkbox
                    )
                    item.setCheckable(checkable)

                    def slot(
                            datum,
                            item=item,
                            field_name=field_name,
                            editable=editable,
                    ):
                        node = item.data(epyqlib.utils.qt.UserRoles.node)
                        model = node.find_root().model
                        field_metadata = getattr(
                            fields(node),
                            field_name,
                        )
                        data_display = field_metadata.data_display

                        field_metadata.updating = True
                        try:
                            display_datum = datum
                            if data_display is not None:
                                display_datum = data_display(
                                    node,
                                    value=display_datum,
                                    model=model,
                                )
                            elif field_metadata.converter == two_state_checkbox:
                                display_datum = ''

                            if display_datum is None:
                                # TODO: CAMPid 0794305784527546542452654254679680
                                # The display role is supposed to be '-' for None
                                # but they can't be different
                                #
                                # http://doc.qt.io/qt-5/qstandarditem.html#data
                                #   The default implementation treats Qt::EditRole
                                #   and Qt::DisplayRole as referring to the same
                                #   data
                                display_text = ''
                                # edit_text = ''
                                if editable:
                                    decoration = QtGui.QColor('green')
                                    decoration.setAlphaF(0.4)
                                else:
                                    decoration = None
                            else:
                                display_text = str(display_datum)
                                # edit_text = display_text
                                decoration = None

                            item.setData(display_text, PyQt5.QtCore.Qt.DisplayRole)
                            # item.setData(edit_text, PyQt5.QtCore.Qt.EditRole)
                            item.setData(decoration, PyQt5.QtCore.Qt.DecorationRole)
                            item.setData(datum, epyqlib.utils.qt.UserRoles.raw)
                            if item.isCheckable():
                                item.setCheckState(
                                    PyQt5.QtCore.Qt.Checked
                                    if datum
                                    else PyQt5.QtCore.Qt.Unchecked,
                                )
                        finally:
                            field_metadata.updating = False

                    connections[getattr(changed_signals, '_pyqtify_signal_' + field_name)] = slot

                    slot(getattr(child, field_name))

                items.append(item)

            parent_item.insertRow(row, items)

        connections[child.pyqt_signals.child_added] = self.child_added
        connections[child.pyqt_signals.child_removed] = self.deleted

        for signal, slot in connections.items():
            signal.connect(slot)

    def pyqtify_disconnect(self, parent, child):
        def visit(node, nodes):
            if node is child:
                this_parent = parent
            else:
                this_parent = node.tree_parent

            nodes.append({'parent': this_parent, 'child': node})

        nodes = []
        child.traverse(call_this=visit, payload=nodes, internal_nodes=True)

        for kwargs in nodes:
            self._pyqtify_disconnect(**kwargs)

    def _pyqtify_disconnect(self, parent, child):
        connection_id = get_connection_id(parent=parent, child=child)
        try:
            connections = self.connected_signals.pop(
                connection_id,
            )
        except KeyError:
            # TODO: why is this even happening?
            return

        for signal, slot in connections.items():
            signal.disconnect(slot)

        self.uuid_to_node.pop(child.uuid)

    def child_added(self, child, row):
        parent = child.tree_parent

        self.pyqtify_connect(parent, child)

        if child.uuid is None:
            check_uuids(self.root)

    def deleted(self, parent, node, row):
        item = self.item_from_node(parent)
        taken_items = item.takeRow(row)
        self.node_to_item.pop(node)
        for taken_item in taken_items:
            try:
                del self._all_items_dict[(
                    taken_item.data(epyqlib.utils.qt.UserRoles.node),
                    taken_item.column(),
                )]
            except KeyError:
                pass
        self.pyqtify_disconnect(parent, node)

    def supportedDropActions(self):
        return QtCore.Qt.MoveAction

    def mimeTypes(self):
        return (self.mime_type,)

    def node_from_index(self, index):
        item = self.model.itemFromIndex(index)
        if item is None:
            node = self.root
        else:
            node = item.data(epyqlib.utils.qt.UserRoles.node)
        return node

    def index_from_node(self, node):
        for root in {self} | self.droppable_from:
            try:
                item = root.item_from_node(node)
            except KeyError:
                continue
            index = root.model.indexFromItem(item)
            return index

        return None

    def mimeData(self, indexes):
        [node] = {self.node_from_index(i) for i in indexes}
        m = QtCore.QMimeData()
        m.setData(self.mime_type, node.uuid.bytes)

        return m

    def dropMimeData(self, data, action, row, column, parent):
        logger.debug('entering dropMimeData()')
        logger.debug((data, action, row, column, parent))

        node, new_parent, source_row = self.source_target_for_drop(
            column, data, parent, row)

        node_to_insert_before = None
        if row != -1:
            node_to_insert_before = new_parent.child_at_row(row)

        if action == QtCore.Qt.MoveAction:
            logger.debug('node name: {}'.format(
                getattr(node, 'name', '<missing attribute>'),
            ))
            logger.debug((data, action, row, column, parent))
            logger.debug('dropped on: {}'.format(
                getattr(new_parent, 'name', '<no name attribute>'),
            ))

            moved = False

            if new_parent is node.tree_parent:
                moved = new_parent.internal_move(
                    node=node,
                    node_to_insert_before=node_to_insert_before,
                )

            if not moved:
                if new_parent.remove_old_on_drop(node=node):
                    node.tree_parent.remove_child(child=node)

                new_child = new_parent.child_from(node=node)

                if new_child is None:
                    pass
                elif row == -1:
                    new_parent.append_child(new_child)
                else:
                    new_row = new_parent.row_of_child(node_to_insert_before)
                    new_parent.insert_child(new_row, new_child)

        # Always returning False so that Qt won't do anything...  like
        # thinking it knows which row of items to delete to finish the
        # move.
        return False

    def source_target_for_drop(self, column, data, parent, row):
        new_parent = self.node_from_index(parent)
        if row == -1 and column == -1:
            if parent.isValid():
                row = 0
            else:
                row = len(self.root.children)
        u = uuid.UUID(bytes=bytes(data.data(self.mime_type)))
        source = self.node_from_uuid(u)
        return source, new_parent, row

    def node_from_uuid(self, u):
        for model in {self} | self.droppable_from:
            node = model.uuid_to_node.get(u)
            if node is not None:
                return node

        raise NotFoundError('''UUID '{}' not found'''.format(u))

    def nodes_from_uuid_list(self, u):
        nodes = []
        for i in u:
            try:
                target_node = self.node_from_uuid(i)
            except NotFoundError:
                target_node = str(u)
            
            nodes.append(target_node)

        return nodes

    def canDropMimeData(self, mime, action, row, column, parent):
        node, new_parent, _ = self.source_target_for_drop(
            column, mime, parent, row)
        can_drop = new_parent.can_drop_on(node=node)

        logger.debug('canDropMimeData: {}: {}, {}'.format(
            getattr(new_parent, 'name', '<no name attribute>'), row, can_drop))

        return can_drop

    def update_nodes(self):
        def visit(node, _):
            update = getattr(node, 'update', None)

            if update is not None:
                update()

        self.root.traverse(call_this=visit, internal_nodes=True)


class Reference(marshmallow.fields.UUID):
    def _serialize(self, value, attr, obj):
        if value is None:
            uuid = None
        else:
            uuid = value.uuid

        return super()._serialize(uuid, attr, obj)


def columns_to_code(c):
    columns = collections.defaultdict(list)

    for x in c:
        columns[x[1]].append(x[0])

    code = []

    for name, types in columns.items():
        if len(types) == 0:
            pass
        elif len(types) == 1:
            code.append(f"(({types[0].__name__}, '{name}'),),")
        else:
            type_code = ', '.join(sorted(cls.__name__ for cls in types))
            type_code = f'({type_code})'
            code.append(f"tuple((x, '{name}') for x in {type_code}),")

    # with this you can copy/paste to fill in the missing columns
    return '\n' + '\n'.join(code)


class Tests:
    def test_all_fields_in_columns(self):
        fields = set()

        for cls in self.types:
            if cls is self.root_type:
                continue

            for field in epyqlib.attrsmodel.fields(cls):
                if field.no_column:
                    continue

                fields.add((cls, field.name))

        columns_list = [
            tuple(x)
            for x in itertools.chain(*(
                column.items()
                for column in self.columns
            ))
            if x[0] is not self.root_type
        ]

        columns = set(columns_list)

        assert len(columns_list) == len(columns)

        extra = columns - fields
        missing = fields - columns

        assert extra == set()
        assert missing == set(), columns_to_code(missing)

    def test_all_have_can_drop_on(self):
        self.assert_incomplete_types(
            name='can_drop_on',
            signature=['node'],
        )

    def test_all_have_can_delete(self):
        self.assert_incomplete_types(
            name='can_delete',
            signature=['node'],
        )

    def test_all_have_remove_old_on_drop(self):
        self.assert_incomplete_types(
            name='remove_old_on_drop',
            signature=['node'],
        )

    def test_all_have_child_from(self):
        self.assert_incomplete_types(
            name='child_from',
            signature=['node'],
        )

    def test_all_have_internal_move(self):
        self.assert_incomplete_types(
            name='internal_move',
            signature=['node', 'node_to_insert_before'],
        )

    def test_all_have_check(self):
        self.assert_incomplete_types(
            name='check',
            signature=['models'],
        )

    def test_all_addable_also_in_types(self):
        # Since addable types is dynamic and could be anything... this
        # admittedly only checks the addable types on default instances.
        for cls in self.types.types.values():
            addable_types = cls.all_addable_types().values()
            assert (
                (set(addable_types) - set(self.types))
                 == set()
             )

    def test_hashability(self):
        expected = []
        bad = []
        for cls in self.types:
            instance = cls()
            try:
                hash(instance)
            except TypeError :
                bad.append(cls)

        sys.stderr.write('\n')
        for cls in bad:
            sys.stderr.write(
                '{path}  {name}\n'.format(
                    path=epyqlib.utils.general.path_and_line(cls),
                    name=cls.__name__,
                ),
            )

        assert bad == expected

    def test_has_uuid(self):
        for cls in self.types:
            assert hasattr(cls, 'uuid')

    def assert_incomplete_types(self, name, signature=None):
        missing = []
        bad_signature = []
        signature = list(signature)

        for cls in self.types.types.values():
            attribute = getattr(cls, name, None)
            if attribute is None:
                missing.append((cls, name))
            elif signature is not None:
                if isinstance(cls.__dict__[name], staticmethod):
                    tweaked_signature = signature
                elif isinstance(cls.__dict__[name], classmethod):
                    tweaked_signature = ['cls', *signature]
                else:
                    tweaked_signature = ['self', *signature]

                actual_signature = inspect.signature(attribute)
                actual_signature = actual_signature.parameters.keys()
                actual_signature = list(actual_signature)
                if tweaked_signature != actual_signature:
                    bad_signature.append(
                        (cls, tweaked_signature, actual_signature),
                    )
                    continue

        sys.stderr.write('\n')
        for cls, attribute in missing:
            sys.stderr.write(
                '{path}  {name} is missing: {attribute}\n'.format(
                    path=epyqlib.utils.general.path_and_line(cls),
                    name=cls,
                    attribute=attribute,
                ),
            )
        for cls, expected, actual in bad_signature:
            sys.stderr.write(
                '{path}  {name} has signature {actual}, should be {expected}\n'.format(
                    path=epyqlib.utils.general.path_and_line(
                        getattr(cls, name)),
                    name=cls,
                    actual=actual,
                    expected=expected,
                ),
            )
        assert [] == missing
        assert [] == bad_signature


def build_tests(types, root_type, columns):
    return type(
        'BuiltTests',
        (Tests,),
        {
            'types': types,
            'root_type': root_type,
            'columns': columns,
        },
    )



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'hmidialog.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(399, 294)
        Form.setProperty("style_small", True)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setContentsMargins(13, 13, 13, 13)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setTextFormat(QtCore.Qt.RichText)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.ok_button = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ok_button.sizePolicy().hasHeightForWidth())
        self.ok_button.setSizePolicy(sizePolicy)
        self.ok_button.setObjectName("ok_button")
        self.horizontalLayout.addWidget(self.ok_button)
        self.cancel_button = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cancel_button.sizePolicy().hasHeightForWidth())
        self.cancel_button.setSizePolicy(sizePolicy)
        self.cancel_button.setObjectName("cancel_button")
        self.horizontalLayout.addWidget(self.cancel_button)
        self.verticalLayout.addLayout(self.horizontalLayout)
        spacerItem = QtWidgets.QSpacerItem(0, 15, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "TextLabel"))
        self.ok_button.setText(_translate("Form", "OK"))
        self.cancel_button.setText(_translate("Form", "Cancel"))



import collections
import enum
import inspect
import itertools
import logging
import math
import os
import pathlib
import shutil
import stat
import traceback
import textwrap
import time
import zipfile

import attr

__copyright__ = 'Copyright 2017, EPC Power Corp.'
__license__ = 'GPLv2+'

logger = logging.getLogger()


class ExpectedException(Exception):
    def expected_message(self):
        return 'This is expected: {}'.format(str(self))


class UnmatchedSerialNumberError(Exception):
    pass


class Container:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


def multiparagraph_wrap(s, *args, **kwargs):
    paragraphs = s.split('\n\n')
    return '\n\n'.join(
        '\n'.join(textwrap.wrap(p, *args, **kwargs))
        for p in paragraphs
    )


@attr.s
class AverageValueRate:
    _seconds = attr.ib(converter=float)
    _deque = attr.ib(default=attr.Factory(collections.deque))

    @attr.s
    class Event:
        time = attr.ib()
        value = attr.ib()
        delta = attr.ib()

    def add(self, value):
        now = time.monotonic()

        if len(self._deque) > 0:
            delta = now - self._deque[-1].time

            cutoff_time = now - self._seconds

            while self._deque[0].time < cutoff_time:
                self._deque.popleft()
        else:
            delta = 0

        event = self.Event(time=now, value=value, delta=delta)
        self._deque.append(event)

    def rate(self):
        if len(self._deque) > 0:
            dv = self._deque[-1].value - self._deque[0].value
            dt = self._deque[-1].time - self._deque[0].time
        else:
            dv = -1
            dt = 0

        if dv <= 0:
            return 0
        elif dt == 0:
            return math.inf

        return dv / dt

    def remaining_time(self, final_value):
        rate = self.rate()
        if rate <= 0:
            return math.inf
        else:
            return (final_value - self._deque[-1].value) / rate


def write_device_to_zip(zip_path, epc_dir, referenced_files, code=None,
                        sha=None, checkout_dir=None):
    # TODO: stdlib zipfile can't create an encrypted .zip
    #       make a good solution that will...
    with zipfile.ZipFile(file=zip_path, mode='w') as zip:
        for device_path in referenced_files:
            filename = os.path.join(epc_dir, device_path)
            zip.write(filename=filename,
                      arcname=os.path.relpath(filename, start=epc_dir))

        if sha is not None:
            sha_file_name = 'sha'
            sha_file_path = os.path.join(checkout_dir, sha_file_name)
            with open(sha_file_path, 'w') as sha_file:
                sha_file.write(sha + '\n')
            zip.write(
                filename=sha_file_path,
                arcname=sha_file_name
            )


# https://docs.python.org/3/library/itertools.html
def pairwise(iterable):
    's -> (s0,s1), (s1,s2), (s2, s3), ...'
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


# https://docs.python.org/3/library/itertools.html
def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


def chunker(iterable, n, factory=tuple):
    """Collect data into fixed-length chunks or blocks except last due to
    lack of fill.
    """
    # chunker('ABCDEFG', 3) --> ABC DEF G"

    i = iter(iterable)
    r = range(n)

    while True:
        # t = factory(next(i) for _ in r)
        t = factory(v for _, v in zip(r, i))

        if len(t) > 0:
            yield t
        else:
            break


def generate_ranges(ids):
    try:
        start = ids[0]
    except IndexError:
        return

    for previous, next in pairwise(itertools.chain(ids, (None,))):
        if previous + 1 != next:
            yield (start, previous)

            start = next


@attr.s
class append_to_method:
    callable = attr.ib()
    name = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.name is None:
            self.name = self.callable.__name__

    def __call__(self, cls):
        old = getattr(cls, self.name, None)
        to_call = self.callable

        def f(s, *args, **kwargs):
            if old is not None:
                old(s, *args, **kwargs)

            to_call(s)

        setattr(cls, self.name, f)
        print('assigned {} to {}.{}'.format(f, cls.__name__, self.name))

        return cls


# Based on https://docs.python.org/3.5/library/enum.html#autonumber
class AutoNumberIntEnum(enum.IntEnum):
    def __new__(cls, value=None):
        if value is None:
            if len(cls.__members__) == 0:
                base = 0
            else:
                base = tuple(cls.__members__.values())[-1]
            value = base + 1

        obj = int.__new__(cls, value)
        obj._value_ = value
        return obj


def enumerated_attrs(enumeration, **kwargs):
    def inner(cls):
        for enumerator in enumeration:
            setattr(cls, enumerator.name, attr.ib(**kwargs))

        cls.enumeration = enumeration

        return cls
    return inner


class TextTable:
    def __init__(self, format_string=None):
        self.format = format_string
        self.rows = []
        self.widths = ()

    def append(self, *row):
        row = tuple(str(c) for c in row)
        self.rows.append(row)
        widths = tuple(len(c) for c in row)
        self.widths = tuple(
            max(a, b) for a, b
            in itertools.zip_longest(self.widths, widths, fillvalue=0)
        )

    def extend(self, rows):
        for row in rows:
            self.append(*row)

    def __str__(self):
        if len(self.rows) == 0:
            return ''

        f = self.format
        if f is None:
            f = ' '.join(('{{:{}}}',) * len(self.widths))

        f = f.format(*(max(w, 1) for w in self.widths))

        return '\n'.join((
            f.format(*(row + (('',) * (len(self.widths) - len(row)))))
            for row in self.rows
        ))


@attr.s
class Collector(list):
    cls = attr.ib()
    payload_attribute = attr.ib()

    def append(self, *args, **kwargs):
        s = super()

        def f(payload):
            payload_dict = {self.payload_attribute: payload}
            s.append(self.cls(
                *args,
                **kwargs,
                **{self.payload_attribute: payload},
            ))

            return payload

        return f


def exception_logger(excType, excValue, tracebackobj):
    logger.error('Uncaught exception hooked:\n{}'.format(
         ''.join(traceback.format_exception(excType, excValue, tracebackobj))
    ))


@attr.s
class TypeMap:
    types = attr.ib(default=attr.Factory(dict), init=False)

    def __call__(self, wrapped):
        def inner(cls):
            self.types[wrapped] = cls

            return cls

        return inner

    def __getitem__(self, item):
        return self.types[item]

    def wrap(self, wrapped, *args, **kwargs):
        return self.types[type(wrapped)](wrapped=wrapped, *args, **kwargs)


def spaced_to_camel(name, upper):
    segments = name.split(' ')

    if not upper:
        segments[0] = segments[0].lower()

    segments = itertools.chain(
        segments[:1],
        *(''.join(itertools.chain(
            c[:1].upper(), c[1:],
        )) for c in segments[1:]),
    )

    return ''.join(segments)


def spaced_to_lower_camel(name):
    return spaced_to_camel(name=name, upper=False)


def spaced_to_upper_camel(name):
    return spaced_to_camel(name=name, upper=True)


def underscored_to_upper_camel(name):
    return ''.join(w.title() for w in name.split('_'))


def underscored_camel_to_upper_camel(name):
    return ''.join(w[:1].upper() + w[1:] for w in name.split('_'))


def cameled_to_spaced(name):
    word_indexes = sorted(
        (0, len(name))
        + tuple(
            i
            for i, (c, next_c) in enumerate(pairwise(name))
            if c.isupper() and next_c.islower()
        )
        + tuple(
            i + 1
            for i, (prev_c, c) in enumerate(pairwise(name))
            if not prev_c.isupper() and c.isupper()
        )
    )

    words = (
        name[start:end]
        for start, end in pairwise(word_indexes)
        if end > start
    )

    return ' '.join(word.strip() for word in words).strip()


def underscored_camel_to_title_spaced(name):
    result = underscored_camel_to_upper_camel(name)
    result = cameled_to_spaced(result)

    return result


def identifier_path(it):
    return '__' + '_'.join(
        it.__module__.split('.') + [it.__qualname__]
    )


# TODO: CAMPid 0238493420143087667542054268097120437916848
# http://stackoverflow.com/a/21263493/228539
def del_rw(action, name, exc):
    os.chmod(name, stat.S_IWRITE)
    if os.path.isdir(name):
        os.rmdir(name)
    else:
        os.remove(name)


def rmtree(path, ignore_errors=False):
    try:
        shutil.rmtree(path, onerror=del_rw)
    except FileNotFoundError:
        if not ignore_errors:
            raise


def contiguous_groups(s):
    groups = []
    group = []

    for previous, current in pairwise(itertools.chain(s, (math.inf,))):
        group.append(previous)
        if current > previous + 1:
            groups.append(group)
            group = []

    return groups


def path_and_line(o):
    return '{}:{}'.format(
        inspect.getsourcefile(o),
        inspect.getsourcelines(o)[1],
    )


def intersect_dicts(a, b):
    common_keys = set(a).intersection(b)
    return {k: a[k] for k in common_keys}


# TODO: CAMPid 073407143081341008467657184603164130
def format_nested_lists(it, indent='    ', indentation=''):
    result = []

    for item in it:
        if isinstance(item, list):
            result.extend(format_nested_lists(
                it=item,
                indent=indent,
                indentation=indentation + indent),
            )
        elif item.strip() == '':
            result.append('')
        else:
            result.append(indentation + item)

    if indentation == '':
        result.append('')
        return '\n'.join(result)
    else:
        return result


def ordered_unique(iterable):
    seen = set()

    for item in iterable:
        if item not in seen:
            seen.add(item)
            yield item


@attr.s
class Orderer:
    ordered = attr.ib()
    indexes = attr.ib()
    key = attr.ib()

    @classmethod
    def build(cls, ordered, key=lambda item: item):
        indexes = {
            item: index
            for index, item in enumerate(ordered)
        }

        return cls(
            ordered=ordered,
            indexes=indexes,
            key=key,
        )

    def __call__(self, value):
        return self.indexes.get(self.key(value), math.inf)



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

import functools

# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class AbstractColumns:
    def __init__(self, **kwargs):
        for member in self._members:
            try:
                value = kwargs[member]
            except KeyError:
                value = None
            finally:
                setattr(self, member, value)

        object.__setattr__(self, '_length', len(self.__dict__))

        invalid_parameters = set(kwargs.keys()) - set(self.__dict__.keys())
        if len(invalid_parameters):
            raise ValueError('Invalid parameter{} passed: {}'.format(
                's' if len(invalid_parameters) > 1 else '',
                ', '.join(invalid_parameters)))

    @classmethod
    def __len__(cls):
        return len(cls._members)

    @classmethod
    def indexes(cls):
        return cls(**dict(zip(cls._members, range(len(cls._members)))))

    @classmethod
    def fill(cls, value):
        return cls(**dict(zip(cls._members, [value] * len(cls._members))))

    @classmethod
    def as_title_case(cls):
        return cls(**dict(zip(cls._members, (s.title() for s in cls._members))))

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    @functools.lru_cache(maxsize=None)
    def index_from_attribute(self, index):
        for attribute in self.__class__._members:
            if index == getattr(self.__class__.indexes, attribute):
                return attribute

        raise IndexError('column index out of range')

    def __getitem__(self, index):
        if index < 0:
            index += len(self)
        return getattr(self, self.index_from_attribute(index))

    def __setitem__(self, index, value):
        if index < 0:
            index += len(self)
        return setattr(self, self.index_from_attribute(index), value)

    def __setattr__(self, name, value):
        if name in self._members:
            object.__setattr__(self, name, value)
        else:
            raise TypeError("Attempted to set attribute {}"
                            .format(name))


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



# This file has been generated


build_system = 'AppVeyor'
build_id = '29367322'
build_number = '402'
build_version = '1.0.402'
job_id = 't6aq8tvtvhs78aus'
job_url = 'https://ci.appveyor.com/project/KyleAltendorf/stlib/build/job/t6aq8tvtvhs78aus'



import os
import pkg_resources
import sys
import textwrap

import click


@click.command()
@click.option('--dist', '--distribution', '-d', 'distribution', required=True)
@click.option('--out', '-o', type=click.File('w'), default=sys.stdout)
def cli(distribution, out):
    import pkg_resources

    version = pkg_resources.get_distribution(distribution).version

    out.write(textwrap.dedent('''\
        # This file is automatically generated by {me}
        __version__ = '{version}'
        __sha__ = '{sha}'
        __revision__ = '{revision}'
    '''.format(
        me=os.path.splitext(os.path.basename(sys.argv[0]))[0],
        version=version,
        sha='',
        revision='',
    )))



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

import weakref

from epyqlib.treenode import TreeNode
import epyqlib.utils.qt
from PyQt5.QtCore import (Qt, QAbstractItemModel, QVariant,
                          QModelIndex, pyqtSignal, pyqtSlot)

# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class PyQAbstractItemModel(QAbstractItemModel):
    root_changed = pyqtSignal(TreeNode)

    def __init__(self, root, checkbox_columns=None, editable_columns=None,
                 alignment=None, parent=None):
        QAbstractItemModel.__init__(self, parent=parent)

        self.root = root
        self.checkbox_columns = checkbox_columns
        self.editable_columns = editable_columns

        if alignment is not None:
            self.alignment = alignment
        else:
            self.alignment = Qt.AlignTop | Qt.AlignLeft


        self.index_from_node_cache = weakref.WeakKeyDictionary()

        self.role_functions = {
            Qt.DisplayRole: self.data_display,
            epyqlib.utils.qt.UserRoles.sort: self.data_display,
            epyqlib.utils.qt.UserRoles.unique: self.data_unique,
            epyqlib.utils.qt.UserRoles.raw: self.data_display,
            Qt.TextAlignmentRole: lambda index: int(self.alignment),
            Qt.CheckStateRole: self.data_check_state,
            Qt.EditRole: self.data_edit,
            Qt.BackgroundRole: self.data_background,
            Qt.ForegroundRole: self.data_foreground,
            Qt.DecorationRole: self.data_decoration,
            Qt.ToolTipRole: self.data_tool_tip,
            Qt.FontRole: self.data_font,
        }

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return QVariant(self.headers[section])
        return QVariant()

    def data_display(self, index):
        node = index.internalPointer()

        column = index.column()
        if column < len(node.fields):
            return node.fields[column]

        return None

    def data_unique(self, index):
        return index.internalPointer().unique()

    def data_check_state(self, index):
        if self.checkbox_columns is not None:
            if self.checkbox_columns[index.column()]:
                node = index.internalPointer()
                if hasattr(node, 'checked'):
                    return node.checked(index.column())

                return None

    def data_edit(self, index):
        node = index.internalPointer()
        if hasattr(node, 'get_human_value'):
            if callable(node.get_human_value):
                value = node.get_human_value(column=index.column())
            else:
                value = ''
        else:
            value = node.fields[index.column()]

        if value is None:
            value = ''
        else:
            value = str(value)

        return value

    def data_background(self, index):
        return None

    def data_foreground(self, index):
        return None

    def data_decoration(self, index):
        return None

    def data_tool_tip(self, index):
        return None

    def data_font(self, index):
        return None

    def data(self, index, role):
        if not index.isValid():
            return None

        if role in self.role_functions:
            return self.role_functions[role](index=index)

        return None

    def flags(self, index):
        flags = QAbstractItemModel.flags(self, index)

        if not index.isValid():
            return flags

        column = index.column()

        if self.editable_columns is not None:
            if self.editable_columns[column]:
                flags |= Qt.ItemIsEditable

        if self.checkbox_columns is not None:
            if self.checkbox_columns[column]:
                flags |= Qt.ItemIsUserCheckable

        return flags

    def index(self, row, column, parent=None):
        if parent is None:
            parent = QModelIndex()

        # TODO: commented out stuff ought to be good rather than
        #       breaking stuff.
        #
        #       http://stackoverflow.com/questions/26680168/pyqt-treeview-index-error-removing-last-row

        # if not self.hasIndex(row, column, parent):
        #     return QModelIndex()

        # if not parent.isValid():
        #     return QModelIndex()

        if row < 0 or column < 0:
            return QModelIndex()

        node = self.node_from_index(parent)
        child = node.child_at_row(row)

        if child is None:
            return QModelIndex()

        return self.createIndex(row, column, child)

    def columnCount(self, parent):
        return len(self.headers)

    def rowCount(self, parent=None):
        if parent is None:
            parent = QModelIndex()

        if parent.column() > 0:
            return 0

        node = self.node_from_index(parent)
        if node is None:
            return 0
        return len(node)

    def parent(self, child):
        if not child.isValid():
            return QModelIndex()

        node = self.node_from_index(child)

        if node is None:
            return QModelIndex()

        parent = node.tree_parent

        if parent in [None, self.root]:
            return QModelIndex()

        grandparent = parent.tree_parent
        if grandparent is None:
            return QModelIndex()
        row = grandparent.row_of_child(parent)

        if row == -1:
            raise Exception('row == -1')

        return self.createIndex(row, 0, parent)

    def node_from_index(self, index):
        if index.isValid():
            return index.internalPointer()
        else:
            return self.root

    def index_from_node(self, node):
        # TODO  make up another role for identification?
        if node in self.index_from_node_cache:
            index = self.index_from_node_cache[node]
        else:
            if node is self.root:
                index = QModelIndex()
            else:
                index = self.index(
                    row=node.tree_parent.row_of_child(node),
                    column=0,
                    parent=self.index_from_node(node.tree_parent)
                )

                self.index_from_node_cache[node] = index

        return index

    @pyqtSlot(TreeNode, int, TreeNode, int, list)
    def changed(self, start_node, start_column, end_node, end_column, roles):
        start_index = self.index_from_node(start_node)
        start_row = start_index.row()
        start_parent = start_index.parent()
        start_index = self.index(start_row, start_column, start_parent)
        if end_node is not start_node:
            end_index = self.index_from_node(end_node)
            end_row = end_index.row()
            end_parent = end_index.parent()
            end_index = self.index(end_row, end_column, end_parent)
        else:
            end_row = start_row
            end_parent = start_parent
            if end_column != start_column:
                end_index = self.index(end_row, end_column, end_parent)
            else:
                end_index = start_index

        self.dataChanged.emit(start_index, end_index, roles)

    @pyqtSlot(TreeNode, int, int)
    def begin_insert_rows(self, parent, start_row, end_row):
        self.beginInsertRows(self.index_from_node(parent), start_row, end_row)

    @pyqtSlot()
    def end_insert_rows(self):
        self.index_from_node_cache.clear()
        self.endInsertRows()

    @pyqtSlot(TreeNode, int, int)
    def begin_remove_rows(self, parent, start_row, end_row):
        self.beginRemoveRows(self.index_from_node(parent), start_row, end_row)

    @pyqtSlot()
    def end_remove_rows(self):
        self.index_from_node_cache.clear()
        self.endRemoveRows()

    @pyqtSlot()
    def set_root(self, root):
        self.beginResetModel()
        self.root = root
        self.endResetModel()
        self.root_changed.emit(root)


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



import contextlib
import functools
import json
import operator
import pathlib
import time
import uuid

import attr
import canmatrix
import twisted.internet.defer

import epyqlib.canneo
import epyqlib.device
import epyqlib.nv
import epyqlib.utils.qt
import epyqlib.utils.twisted
import epyqlib.utils.units
import epyqlib.updateepc


class FormatVersionError(Exception):
    pass


class AlreadyLoadedError(Exception):
    pass


class BusAlreadySetError(Exception):
    pass


supported_version = [2]


def format_version_validator(instance, attribute, value):
    if value != supported_version:
        raise FormatVersionError(
            'Only format_version {} is supported'.format(
                '.'.join(str(v) for v in supported_version),
            ),
        )


@attr.s
class Definition:
    base_path = attr.ib()
    format_version = attr.ib(
        validator=format_version_validator,
    )
    can_path = attr.ib(converter=pathlib.Path)
    # can_configuration = attr.ib()
    nv_configuration = attr.ib()
    node_id_type = attr.ib()
    access_level_path = attr.ib()
    access_password_path = attr.ib()
    node_id = attr.ib(default=247)
    controller_id = attr.ib(default=65)
    # nv_range_check_overridable = attr.ib()
    # node_id_type = attr.ib()
    # name = attr.ib()
    # tabs = attr.ib()
    # ui_paths = attr.ib()
    # module = attr.ib()
    # parameter_hierarchy = attr.ib()
    # nv_meta_enum = attr.ib()

    def load_can(self):
        matrix, = canmatrix.formats.loadp(
            str(self.base_path / self.can_path)
        ).values()

        return matrix

    @classmethod
    def load(cls, file, base_path=None):
        if base_path is None:
            base_path = pathlib.Path(file.name).parents[0]

        return cls.loads(s=file.read(), base_path=base_path)

    @classmethod
    def loadp(cls, path):
        with open(path) as f:
            return cls.load(f)

    @classmethod
    def loads(cls, s, base_path):
        raw = json.loads(s)

        access_level_path = raw.get('access_level_path')
        if access_level_path is not None:
            access_level_path = access_level_path.split(';')

        access_password_path = raw.get('access_password_path')
        if access_password_path is not None:
            access_password_path = access_password_path.split(';')

        return Definition(
            base_path=base_path,
            format_version=raw['format_version'],
            can_path=base_path / raw['can_path'],
            nv_configuration=raw['nv_configuration'],
            node_id=raw.get('node_id'),
            node_id_type=raw.get(
                'node_id_type',
                next(iter(epyqlib.device.node_id_types)),
            ),
            controller_id=raw.get('controller_id'),
            access_level_path=access_level_path,
            access_password_path=access_password_path,
        )


operator_map = {
    '<': operator.lt,
    '<=': operator.le,
    '==': operator.eq,
    '!=': operator.ne,
    '>=': operator.ge,
    '>': operator.gt,
}

reverse_operator_map = {
    v: k
    for k, v in operator_map.items()
}


@attr.s
class Signal:
    signal = attr.ib()
    device = attr.ib()

    @twisted.internet.defer.inlineCallbacks
    def get(self, stale_after=0.1, timeout=1):
        start = time.time()

        last_received = self.signal.last_received()

        if last_received is None or last_received <= start - stale_after:
            yield epyqlib.utils.qt.signal_as_deferred(
                self.signal.value_set,
                timeout=timeout,
            )

        value = self.signal.to_human(self.signal.value)

        if value in self.signal.enumeration:
            return self.signal.enumeration[value]

        return value * self.units()

    def set(self, value):
        units = self.units()
        if units != epyqlib.utils.units.registry.dimensionless:
            value = value.to(units).magnitude

        self.signal.set_human_value(value)
        self.signal.frame._send(update=True)

    def units(self):
        return epyqlib.utils.units.registry.parse_expression(
            self.signal.unit,
        )

    def cyclic_send(self, period):
        self.device.cyclic_send_signal(self, period=period)

    @twisted.internet.defer.inlineCallbacks
    def wait_for(self, op, value, timeout):
        op = operator_map.get(op, op)
        operator_string = reverse_operator_map.get(op, str(op))

        def check():
            present_value = yield self.get()
            return op(present_value, value)

        yield epyqlib.utils.twisted.wait_for(
            check=check,
            timeout=timeout,
            message=(
                f'{self.signal.name} not {operator_string} {value} '
                f'within {timeout:.1f} seconds'
            ),
        )

    def scaling_factor(self):
        return self.signal.factor

    def decimal_places(self):
        return self.signal.get_decimal_places()

    def f_string(self):
        return f'.{self.decimal_places()}f'


@attr.s
class Nv:
    nv = attr.ib()
    device = attr.ib()

    async def set(
            self,
            value=None,
            user_default=None,
            factory_default=None,
            minimum=None,
            maximum=None,
    ):
        values = {
            epyqlib.nv.MetaEnum.maximum: maximum,
            epyqlib.nv.MetaEnum.minimum: minimum,
            epyqlib.nv.MetaEnum.factory_default: factory_default,
            epyqlib.nv.MetaEnum.user_default: user_default,
            epyqlib.nv.MetaEnum.value: value,
        }

        values = {
            meta: value
            for meta, value in values.items()
            if value is not None
        }

        for meta, value in values.items():
            await self.set_meta(value=value, meta=meta)

    async def set_meta(self, value, meta):
        if meta == epyqlib.nv.MetaEnum.value:
            self.nv.set_value(value)
        else:
            getattr(self.nv.meta, meta.name).set_value(value)

        # TODO: verify value was accepted
        await self.device.nvs.protocol.write(
            nv_signal=self.nv,
            meta=meta,
        )

    @contextlib.asynccontextmanager
    async def temporary_set(
            self,
            value=None,
            user_default=None,
            factory_default=None,
            minimum=None,
            maximum=None,
            read_context=None,
            set_context=None,
            restoration_context=None,
    ):
        @contextlib.asynccontextmanager
        async def async_null_context(enter_result=None):
            with contextlib.nullcontext(enter_result=enter_result) as result:
                yield result

        if read_context is None:
            read_context = async_null_context
        if set_context is None:
            set_context = async_null_context
        if restoration_context is None:
            restoration_context = async_null_context

        values = {
            epyqlib.nv.MetaEnum.maximum: maximum,
            epyqlib.nv.MetaEnum.minimum: minimum,
            epyqlib.nv.MetaEnum.factory_default: factory_default,
            epyqlib.nv.MetaEnum.user_default: user_default,
            epyqlib.nv.MetaEnum.value: value,
        }

        values = {
            meta: value
            for meta, value in values.items()
            if value is not None
        }

        original = {}

        try:
            async with read_context():
                for meta, value in values.items():
                    original[meta] = await self.get(meta=meta)

            async with set_context():
                for meta, value in values.items():
                    await self.set_meta(value=value, meta=meta)

            yield
        finally:
            async with restoration_context():
                for meta, value in original.items():
                    await self.set_meta(value=value, meta=meta)

    @twisted.internet.defer.inlineCallbacks
    def get(self, meta=epyqlib.nv.MetaEnum.value):
        value, _meta = yield self.device.nvs.protocol.read(
            nv_signal=self.nv,
            meta=meta,
        )

        return value

    @twisted.internet.defer.inlineCallbacks
    def wait_for(self, op, value, timeout, ignore_read_failures=False):
        op = operator_map.get(op, op)
        operator_string = reverse_operator_map.get(op, str(op))

        @twisted.internet.defer.inlineCallbacks
        def check():
            try:
                own_value = yield self.get()
            except epyqlib.twisted.nvs.RequestTimeoutError:
                if ignore_read_failures:
                    return False

                raise

            return op(own_value, value)

        yield epyqlib.utils.twisted.wait_for(
            check=check,
            timeout=timeout,
            message=(
                f'{self.nv.name} not {operator_string} {value} '
                f'within {timeout:.1f} seconds'
            ),
        )

    def parameter_uuid(self):
        return self.nv.parameter_uuid


@attr.s
class Device:
    definition_path = attr.ib(converter=pathlib.Path)
    definition = attr.ib(default=None)
    canmatrix = attr.ib(default=None)
    neo = attr.ib(default=None)
    nvs = attr.ib(default=None)
    bus = attr.ib(default=None)
    cyclic_frames = attr.ib(default=attr.Factory(set))
    default_elevated_access_level = attr.ib(default=None)
    default_access_level_password = attr.ib(default=None)
    save_nv = attr.ib(default=None)
    save_nv_value = attr.ib(default=None)
    uuid = attr.ib(default=uuid.uuid4)

    def load(self):
        if self.definition is not None:
            raise AlreadyLoadedError('The definition has already been loaded')

        with epyqlib.updateepc.updated(self.definition_path) as updated:
            self.definition = Definition.loadp(updated)
            matrix = self.definition.load_can()

        node_id_adjust = functools.partial(
            epyqlib.device.node_id_types[self.definition.node_id_type],
            device_id=self.definition.node_id,
            controller_id=self.definition.controller_id,
        )

        self.neo = epyqlib.canneo.Neo(
            matrix=matrix,
            node_id_adjust=node_id_adjust,
        )

        nv_neo = epyqlib.canneo.Neo(
            matrix=matrix,
            frame_class=epyqlib.nv.Frame,
            signal_class=epyqlib.nv.Nv,
            strip_summary=False,
            node_id_adjust=node_id_adjust,
        )
        self.nvs = epyqlib.nv.Nvs(
            neo=nv_neo,
            configuration=self.definition.nv_configuration,
            access_level_path=self.definition.access_level_path,
            access_password_path=self.definition.access_password_path,
        )

        self.save_nv = self.nv(
            self.nvs.save_frame.mux_name,
            self.nvs.save_signal.name,
        )
        self.save_nv_value = self.nvs.save_value

    def set_bus(self, bus):
        if self.bus is not None:
            raise BusAlreadySetError()

        try:
            self.neo.set_bus(bus=bus)
            self.nvs.set_bus(bus=bus)
        except:
            # TODO: actually rollback a partial setting
            self.bus = object()
            raise

        self.bus = bus
        self.bus.notifier.add(self.neo)
        self.bus.notifier.add(self.nvs)

    # @functools.lru_cache(maxsize=512)
    def signal(self, *path):
        return Signal(
            signal=self.neo.signal_by_path(*path),
            device=self,
        )

    # @functools.lru_cache(maxsize=512)
    def nv(self, *path):
        return Nv(
            nv=self.nvs.signal_from_names(*path),
            device=self,
        )

    @twisted.internet.defer.inlineCallbacks
    def active_to_nv(self, wait=False):
        yield twisted.internet.defer.ensureDeferred(
            self.save_nv.set(value=self.save_nv_value),
        )

        if wait:
            yield self.wait_for_nv_save_completion()

    @twisted.internet.defer.inlineCallbacks
    def wait_for_nv_save_completion(self):
        nv = self.nv('StatusWarnings', 'eeSaveInProgress')

        yield epyqlib.utils.twisted.sleep(2)

        yield nv.wait_for(
            op='==',
            value=0,
            timeout=120,
            ignore_read_failures=True,
        )

    def cyclic_send_signal(self, signal, period):
        frame = signal.signal.frame
        frame.cyclic_request(self.uuid, period)
        if period is not None:
            self.cyclic_frames.add(frame)
        else:
            self.cyclic_frames.discard(frame)

    def cancel_all_cyclic_sends(self):
        for frame in self.cyclic_frames:
            frame.cyclic_request(self.uuid, None)

    @twisted.internet.defer.inlineCallbacks
    def set_access_level(self, level=None, password=None):
        if level is None:
            level = self.default_elevated_access_level

        if password is None:
            password = self.default_access_level_password

        self.nvs.password_node.set_value(password)
        self.nvs.access_level_node.set_value(level)

        selected_nodes = tuple(
            node
            for node in (
                self.nvs.password_node,
                self.nvs.access_level_node,
            )
            if node is not None
        )

        yield self.nvs.write_all_to_device(only_these=selected_nodes)

    @contextlib.asynccontextmanager
    async def temporary_access_level(self, level=None, password=None):
        access_level_parameter = self.nv(*self.definition.access_level_path[1:])
        original_access_level = await access_level_parameter.get()

        try:
            await self.set_access_level(level=level, password=password)
            yield
        finally:
            await self.set_access_level(
                level=original_access_level,
                password=password,
            )



import pathlib

import alqtendpy.compileui

import setuptools
import versioneer


alqtendpy.compileui.compile_ui(
    directory_paths=[pathlib.Path(__file__).parent / 'epyqlib'],
)


setuptools.setup(
    name="epyqlib",
    author="EPC Power Corp.",
    classifiers=[
        ("License :: OSI Approved :: "
         "GNU General Public License v2 or later (GPLv2+)")
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    entry_points={
        'console_scripts': [
            'outparsecheck = epyqlib.outparsecheck:main [dss]',
            'collectdevices = epyqlib.collectdevices:main',
            'contiguouscommits = epyqlib.utils.contiguouscommits:_entry_point [dulwich]',
            'epyqflash = epyqlib.flash:_entry_point',
            'patchvenv = epyqlib.patchvenv:main',
            'cangenmanual = epyqlib.cangenmanual:_entry_point',
            'updateepc = epyqlib.updateepc:main',
            'genbuildinfo = epyqlib.genbuildinfo:write_build_file',
            'versionfile = epyqlib.cli.versionfile:cli',
            'generateversion = epyqlib.cli.generateversion:cli',
            'autodevice = epyqlib.autodevice.cli:cli',
        ],
        'pytest11': [
            'epyqlib = epyqlib.tests.pytest_plugin',
        ]
    },
    install_requires=[
        'alqtendpy',
        'arrow',
        'bitstruct',
        'canmatrix',
        'click>=7',
        'python-dotenv',
        'natsort',
        'pint',
        'pyelftools',
        'qt5reactor',
        'gitpython',
        'graham>=0.1.11',
        'PyQt5',
        'python-docx',
        'python-can',
        'QtAwesome',
        'siphash-cffi',
        'Twisted',
    ],
    extras_require={
        'deploy': [
            'gitpython',
            'requests',
        ],
        'dss': [
            'ccstudiodss>=0.2.7',
        ],
        'dulwich': [
            'dulwich',
        ],
        'test': [
            'pytest',
            'pytest-qt',
            'pytest-rerunfailures',
        ],
    },
)



import pathlib

import attr
import graham
import marshmallow

import epyqlib.attrsmodel
import epyqlib.pm.parametermodel
import epyqlib.treenode


class SaveCancelled(Exception):
    pass


def create_blank(parameter_model=None):
    value_set = ValueSet(parameter_model=parameter_model)
    _post_load(value_set)

    return value_set


def loads(s, path=None):
    root = graham.schema(Root).loads(s).data
    value_set = ValueSet(path=path)

    if path is not None:
        value_set.path = pathlib.Path(path).absolute()

    _post_load(value_set, root=root)

    return value_set


def load(f):
    value_set = loads(f.read(), path=f.name)

    return value_set


def loadp(path):
    with open(path) as f:
        return load(f)


def _post_load(value_set, root=None):
    if root is None:
        root = Root()

    if value_set.model is None:
        value_set.model = epyqlib.attrsmodel.Model(
            root=root,
            columns=columns,
        )


def copy_parameter_data(
        value_set,
        human_names=True,
        base_node=None,
        calculate_unspecified_min_max=False,
        can_root=None,
):
    def traverse(node, _):
        if isinstance(node, epyqlib.pm.parametermodel.Parameter):
            if human_names:
                name = node.name
            else:
                name = ':'.join((
                    node.original_multiplexer_name,
                    node.original_signal_name,
                ))

            minimum = node.minimum
            maximum = node.maximum

            if calculate_unspecified_min_max and None in (minimum, maximum):
                query_multiplexed_message, = can_root.nodes_by_attribute(
                    attribute_value='Parameter Query',
                    attribute_name='name',
                )
                signal, = query_multiplexed_message.nodes_by_attribute(
                    attribute_value=node.uuid,
                    attribute_name='parameter_uuid',
                )

                calculated_minimum, calculated_maximum = (
                     signal.calculated_min_max()
                )

                if minimum is None:
                    minimum = calculated_minimum

                if maximum is None:
                    maximum = calculated_maximum

            value_set.model.root.append_child(
                Parameter(
                    name=name,
                    parameter_uuid=node.uuid,
                    factory_default=node.default,
                    minimum=minimum,
                    maximum=maximum,
                ),
            )

            value_set.model.root.children.sort()

    if base_node is None:
        base_node = value_set.parameter_model.root

    base_node.traverse(
        call_this=traverse,
        internal_nodes=False,
    )


def decimal_attrib(**kwargs):
    attrib = attr.ib(
        converter=epyqlib.attrsmodel.to_decimal_or_none,
        **kwargs,
    )
    graham.attrib(
        attribute=attrib,
        field=marshmallow.fields.Decimal(
            allow_none=kwargs.get('default', False) is None,
            as_string=True,
        ),
    )
    
    return attrib


@graham.schemify(tag='parameter')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Parameter(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default='New Parameter',
        converter=str,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )

    value = decimal_attrib(default=None)
    user_default = decimal_attrib(default=None)
    factory_default = decimal_attrib(default=None)
    minimum = decimal_attrib(default=None)
    maximum = decimal_attrib(default=None)

    parameter_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
    )

    epyqlib.attrsmodel.attrib(
        attribute=parameter_uuid,
        human_name='Parameter UUID',
    )

    readable = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(allow_none=True),
        ),
    )
    writable = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(allow_none=True),
        ),
    )

    uuid = epyqlib.attrsmodel.attr_uuid(load_only=True)

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    can_delete = epyqlib.attrsmodel.childless_can_delete
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


Root = epyqlib.attrsmodel.Root(
    default_name='Value Set',
    valid_types=(Parameter,),
)

types = epyqlib.attrsmodel.Types(
    types=(
        Root,
        Parameter,
    ),
)


@attr.s
class ValueSet:
    parameter_model = attr.ib(default=None)
    model = attr.ib(default=None)
    path = attr.ib(default=None)
    filters = attr.ib(
        default=(
            ('Parameter Value Set', ['pmvs']),
            ('All Files', ['*']),
        )
    )

    def save(self, path=None, parent=None):
        if path is None:
            path = self.path

        if path is None:
            path = epyqlib.utils.qt.file_dialog(
                filters=self.filters,
                parent=parent,
                save=True,
                caption='Save Value Set As',
            )

            if path is None:
                raise SaveCancelled()

            self.path = pathlib.Path(path)

        sorted_children = sorted(self.model.root.children)

        sorted_root = attr.evolve(
            self.model.root,
            children=sorted_children
        )

        s = graham.dumps(sorted_root, indent=4).data

        with open(path, 'w') as f:
            f.write(s)

            if not s.endswith('\n'):
                f.write('\n')


# TODO: CAMPid 943896754217967154269254167
def merge(name, *types):
    return tuple((x, name) for x in types)


columns = epyqlib.attrsmodel.columns(
    merge('name', *types.types.values()),
    merge('value', Parameter),
    merge('user_default', Parameter),
    merge('factory_default', Parameter),
    merge('minimum', Parameter),
    merge('maximum', Parameter),
    merge('readable', Parameter),
    merge('writable', Parameter),
    merge('parameter_uuid', Parameter),
    merge('uuid', *types.types.values()),
)



import epyqlib.utils.general


def test_container():
    c = epyqlib.utils.general.Container(a=1, b=2)

    assert hasattr(c, 'a')
    assert hasattr(c, 'b')
    assert c.a == 1
    assert c.b == 2


def test_spaced_to_lower_camel():
    name = 'ABC Def ghi LM1'
    expected = 'abcDefGhiLM1'

    result = epyqlib.utils.general.spaced_to_lower_camel(name)

    assert result == expected


def test_spaced_to_upper_camel():
    name = 'ABC Def ghi LM1'
    expected = 'ABCDefGhiLM1'

    result = epyqlib.utils.general.spaced_to_upper_camel(name)

    assert result == expected


def test_underscored_to_upper_camel():
    name = 'abc_de_fgh_IJK'
    expected = 'AbcDeFghIjk'

    result = epyqlib.utils.general.underscored_to_upper_camel(name)

    assert result == expected


def test_underscored_camel_to_upper_camel():
    name = 'abc_de_fgh_IJK'
    expected = 'AbcDeFghIJK'

    result = epyqlib.utils.general.underscored_camel_to_upper_camel(name)

    assert result == expected


def test_cameled_to_spaced():
    name = 'abcDeFghIjk - lmnOp'
    expected = 'abc De Fgh Ijk - lmn Op'

    result = epyqlib.utils.general.cameled_to_spaced(name)

    assert result == expected


def test_cameled_to_spaced_dashed():
    name = 'ABC - DefGhi'
    expected = 'ABC - Def Ghi'

    result = epyqlib.utils.general.cameled_to_spaced(name)

    assert result == expected


def test_cameled_to_spaced_acronym():
    name = ' XY1ABCDefGhiJKLMnOP1 '
    expected = 'XY1 ABC Def Ghi JKL Mn OP1'

    result = epyqlib.utils.general.cameled_to_spaced(name)

    assert result == expected


def test_underscored_camel_to_title_spaced():
    name = 'abcDeFghIjk_lmnOp'
    expected = 'Abc De Fgh Ijk Lmn Op'

    result = epyqlib.utils.general.underscored_camel_to_title_spaced(name)

    assert result == expected


def test_underscored_camel_to_title_spaced_acronym():
    name = 'XY1ABCDefGhi_lmnOP1'
    expected = 'XY1 ABC Def Ghi Lmn OP1'

    result = epyqlib.utils.general.underscored_camel_to_title_spaced(name)

    assert result == expected


def test_underscored_camel_to_title_spaced_acronym_with_number_alone():
    name = 'AB1'
    expected = 'AB1'

    result = epyqlib.utils.general.underscored_camel_to_title_spaced(name)

    assert result == expected
